import os
import sys
import json
from pystrict import strict
from typing import List, Set, Dict, Tuple, Optional, Any, Union, NewType
from urllib.parse import quote_plus

path = os.path.abspath(os.path.dirname(__file__))
(head, tail) = os.path.split(path)
if not head in sys.path:
    sys.path.insert(0, head)
if not path in sys.path:
    sys.path.insert(0, path)

from .helpers import Actions, BaseError
from .langstring import Languages
from .connection import Connection
from .model import Model
from .group import Group
from .project import Project

"""
This module implements the handling (CRUD) of Knora users.

CREATE:
    * Instantiate a new object of the class User with all required parameters
    * Call the ``create``-method on the instance to create the new user

READ:
    * Instantiate a new objects with ``id``(IRI of user) given
    * Call the ``read``-method on the instance
    * Access the information that has been ptovided to the instance

UPDATE:
    * You need an instance of an existing User by reading an instance
    * Change the attributes by assigning the new values
    * Call the ``update```method on the instance

DELETE
    * Instantiate a new objects with ``id``(IRI of user) given, or use any instance that has the id set
    * Call the ``delete``-method on the instance

In addition there is a static methods ``getAllProjects`` which returns a list of all projects
"""

@strict
class User(Model):
    """
    This class represents a user in Knora.

    Attributes
    ----------

    id : str
        IRI of the user [readonly, cannot be modified after creation of instance]

    username : str
        Unique identifier (not an IRI) of the user [read/write]

    email : str
        Email of the user [read/write]

    givenName : str
        Given name (firstname) of the user [read/write]

    familyName : str
        Family name of user (lastname) [read/write]

    password : str
        Password of user [write only]

    lang : Language
        Preferred language of the user. For setting can be Language instance or string "EN", "DE", "FR", "IT"

    status : bool
        Status of the user, If active, is set to True, otherwise false [read/write]

    sysadmin : bool
        True, if user is system administrator [read/write]

    in_groups : Set[str]
        Set of group IRI's the user is member of [readonly].
        Use ``addToGroup``and ``rmFromGroup`` to modify group membership

    in_projects : Set[str]
        Set of project IRI's the user belongs to
        Use ``addToproject``and ``rmFromproject`` to modify project membership


    Methods
    -------

    create : Knora user information object
        Creates a new user and returns the information about this user as it is in Knora

    read : Knora user information object
        Read user data

    update : Knora user information object
        Updates the changed attributes of a user and returns the updated information  as it is in Knora

    delete : Knora result code
        Deletes a user and returns the result code

    addToGroup : None
        Add the user to the given group (will be executed when calling ``update``)

    rmFromGroup : None
        Remove a user from a group (will be executed when calling ``update``)

    addToProject : None
        adds a user to a project, optional as project administrator (will be executed when calling ``update``)

    rmFromProject : None
        removes a user from a project (will be executed when calling ``update``)

    makeProjectAdmin : None
        Promote user to project admin of given project

    unmakeProjectAdmin : None
        Revoke project admin flog for user for given project

    getAllUsers : list of user
        Get a list of all users

    print : None
        Prints the user information to stdout


    """

    _id: str
    _username: str
    _email: str
    _givenName: str
    _familyName: str
    _password: str
    _lang: Languages
    _status: bool
    _sysadmin: bool
    _in_groups: Set[str]
    _in_projects: Dict[str, bool]
    _add_to_project: Dict[str, bool]
    _rm_from_project: Dict[str, bool]
    _add_to_group: Set[str]
    _rm_from_group: Set[str]
    _change_admin: Set[str]

    def __init__(self,
                 con:  Connection,
                 id: Optional[str] = None,
                 username: Optional[str] = None,
                 email: Optional[str] = None,
                 givenName: Optional[str] = None,
                 familyName: Optional[str] = None,
                 password: Optional[str] = None,
                 lang: Optional[Union[str,Languages]] = None,
                 status: Optional[bool] = None,
                 sysadmin: Optional[bool] = None,
                 in_projects: Optional[Dict[str, bool]] = None,
                 in_groups: Optional[Set[str]] = None):
        """
        Constructor for User

        The constructor is user internally or externally, when a new user should be created in Knora.

        :param con: Connection instance [required]
        :param id: IRI of the user [required for CREATE, READ]
        :param username: Username [required for CREATE]
        :param email: Email address [required for CREATE]
        :param givenName: Given name (firstname) of user [required for CREATE]
        :param familyName: Family name (lastname) of user [required for CREATE]
        :param password: Password [required for CREATE]
        :param lang: Preferred language of the user [optional]
        :param status: Status (active = True, inactive/deleted = False) [optional]
        :param sysadmin: User has system administration privileges [optional]
        :param in_projects: Dict with project-IRI as key, boolean(True=project admin) as value [optional]
        :param in_groups: Set with group-IRI's the user should belong to [optional]
        """
        super().__init__(con)
        self._id = str(id) if id is not None else None
        self._username = str(username) if username is not None else None
        self._email = str(email) if email is not None else None
        self._givenName = str(givenName) if givenName is not None else None
        self._familyName = str(familyName) if familyName is not None else None
        self._password = str(password) if password is not None else None
        if lang is not None:
            if isinstance(lang, Languages):
                self._lang = lang
            else:
                lmap = dict(map(lambda a: (a.value, a), Languages))
                if lmap.get(lang) is None:
                    raise BaseError('Invalid language string "' + lang  + '"!')
                self._lang =  lmap[lang]
        else:
            self._lang = None
        self._status = None if status is None else bool(status)

        if in_projects is None or isinstance(in_projects, dict):
            self._in_projects = in_projects if in_projects is not None else {}
        else:
            raise BaseError("In_project must be tuple (project-iri: str, is-admin: bool)!")

        if in_groups is None or isinstance(in_groups, set):
            self._in_groups = in_groups if in_groups is not None else set()
        else:
            raise BaseError('In_groups must be a set of strings or None!')

        self._sysadmin = None if sysadmin is None else bool(sysadmin)
        self._add_to_project = {}
        self._rm_from_project = {}
        self._change_admin = {}
        self._add_to_group = set()
        self._rm_from_group = set()

    @property
    def id(self) -> Optional[str]:
        return self._id

    @id.setter
    def id(self, value: str) -> None:
        raise BaseError('User id cannot be modified!')

    @property
    def username(self) -> Optional[str]:
        return  self._username

    @username.setter
    def username(self, value: Optional[str]):
        if value is None:
            return
        self._username = str(value)
        self._changed.add('username')

    @property
    def email(self) -> Optional[str]:
        return self._email

    @email.setter
    def email(self, value: Optional[str]):
        if value is None:
            return
        self._email = str(value)
        self._changed.add('email')

    @property
    def givenName(self) -> Optional[str]:
        return self._givenName

    @givenName.setter
    def givenName(self, value: Optional[str]):
        if value is None:
            return
        self._givenName = str(value)
        self._changed.add('givenName')

    @property
    def familyName(self) -> Optional[str]:
        return self._familyName

    @familyName.setter
    def familyName(self, value: Optional[str]):
        if value is None:
            return
        self._familyName = str(value)
        self._changed.add('familyName')

    @property
    def password(self) -> Optional[str]:
        return None

    @password.setter
    def password(self, value: Optional[str]):
        if value is None:
            return
        self._password = str(value)
        self._changed.add('password')

    @property
    def lang(self) -> Optional[Languages]:
        return self._lang

    @lang.setter
    def lang(self, value: Optional[Union[str, Languages]]):
        if value is None:
            return
        if isinstance(value, Languages):
            self._lang = value
            self._changed.add('lang')
        else:
            lmap = dict(map(lambda a: (a.value, a), Languages))
            if lmap.get(value) is None:
                raise BaseError('Invalid language string "' + value + '"!')
            self._lang = lmap[value]
            self._changed.add('lang')

    @property
    def status(self) -> bool:
        return self._status

    @status.setter
    def status(self, value: Optional[bool]) -> None:
        self._status = None if value is None else bool(value)
        if value is not None:
            self._changed.add('status')

    @property
    def sysadmin(self) -> bool:
        return self._sysadmin

    @sysadmin.setter
    def sysadmin(self, value: bool):
        self._sysadmin = None if value is None else bool(value)
        if value is not None:
            self._changed.add('sysadmin')

    @property
    def in_groups(self) -> Set[str]:
        return self._in_groups

    @in_groups.setter
    def in_groups(self, value: Any):
        raise BaseError('Group membership cannot be modified directly! Use methods "addToGroup" and "rmFromGroup"')

    def addToGroup(self, value: str):
        """
        Add the user to the given group (executed at next update)

        :param value: IRI of the group
        :return: None
        """

        if value in self._rm_from_group:
            self._rm_from_group.pop(value)
        elif value not in self._in_groups:
            self._add_to_group.add(value)
            self._changed.add('in_groups')
        else:
            raise BaseError("Already member of this group!")

    def rmFromGroup(self, value: str):
        """
        Remove the user from the given group (executed at next update)

        :param value: Group IRI
        :return: None
        """

        if value in self._add_to_group:
            self._add_to_group.discard(value)
        elif value in self._in_groups:
            self._rm_from_group.add(value)
            self._changed.add('in_groups')
        else:
            raise BaseError("User is not in groups!")

    @property
    def in_projects(self) -> Dict[str, bool]:
        return self._in_projects

    @in_projects.setter
    def in_project(self, value: Any):
        raise BaseError('Project membership cannot be modified directly! Use methods "addToProject" and "rmFromProject"')

    def addToProject(self, value: str, padmin: bool = False):
        """
        Add the user to the given project (executed at next update)

        :param value: project IRI
        :param padmin: True, if user should be project admin, False otherwise
        :return: None
        """

        if value in self._rm_from_project:
            self._rm_from_project.pop(value)
        elif value not in self._in_projects:
            self._add_to_project[value] = padmin
            self._changed.add('in_projects')
        else:
            raise BaseError("Already member of this project!")

    def rmFromProject(self, value: str):
        """
        Remove the user from the given project (executed at next update)

        :param value: Project IRI
        :return: None
        """

        if value in self._add_to_project:
            self._add_to_project.pop(value)
        elif value in self._in_projects:
            self._rm_from_project[value] = self._in_projects[value]
            self._changed.add('in_projects')
        else:
            raise BaseError("Project is not in list of member projects!")

    def makeProjectAdmin(self, value: str):
        """
        Make the user project administrator in the given project  (executed at next update)

        :param value: Project IRI
        :return: None
        """

        if value in self._in_projects:
            self._change_admin[value] = True
            self._changed.add('in_projects')
        elif value in self._add_to_project:
            self._add_to_project[value] = True
        else:
            raise BaseError("User is not member of project!")

    def unmakeProjectAdmin(self, value: str):
        """
        Revoke project administrator right for the user from the given project  (executed at next update)

        :param value: Project IRI
        :return: None
        """
        if value in self._in_projects:
            self._change_admin[value] = False
            self._changed.add('in_projects')
        elif value in self._add_to_project:
            self._add_to_project[value] = False
        else:
            raise BaseError("User is not member of project!")

    @property
    def changed(self) -> Set[str]:
        return self._changed

    def has_changed(self, name: str):
        return name in self._changed

    @classmethod
    def fromJsonObj(cls, con: Connection, json_obj: Any):
        """
        Internal method! Should not be used directly!

        This method is used to create a User instance from the JSON data returned by Knora

        :param con: Connection instance
        :param json_obj: JSON data returned by Knora as python3 object
        :return: User instance
        """

        id = json_obj.get('id')
        if id is None:
            raise BaseError('User "id" is missing in JSON from knora')
        email = json_obj.get('email')
        if email is None:
            raise BaseError('User "email" is missing in JSON from knora')
        username = json_obj.get('username')
        if username is None:
            raise BaseError('User "username" is missing in JSON from knora')
        familyName = json_obj.get('familyName')
        givenName = json_obj.get('givenName')
        lang = json_obj.get('lang')
        status = json_obj.get('status')
        if status is None:
            raise BaseError("Status is missing in JSON from knora")

        in_projects: Dict[str, bool] = {}
        in_groups: Set[str] = set()
        if json_obj.get('permissions') is not None and json_obj['permissions'].get('groupsPerProject') is not None:
            sysadmin = False
            project_groups = json_obj['permissions']['groupsPerProject']
            for project in project_groups:
                if project == Project.SYSTEM_PROJECT:
                    if Group.PROJECT_SYSTEMADMIN_GROUP in project_groups[project]:
                        sysadmin = True
                else:
                    for group in project_groups[project]:
                        if group == Group.PROJECT_MEMBER_GROUP:
                            if in_projects.get(project) is None:
                                in_projects[project] = False
                        elif group == Group.PROJECT_ADMIN_GROUP:
                            in_projects[project] = True
                        else:
                            in_groups.add(group)
        return cls(con=con,
                   id=id,
                   username=username,
                   email=email,
                   givenName=givenName,
                   familyName=familyName,
                   lang=lang,
                   status=status,
                   sysadmin=sysadmin,
                   in_projects=in_projects,
                   in_groups=in_groups)

    def toJsonObj(self, action: Actions):
        """
        Internal method! Should not be used directly!

        Creates a JSON-object from the Project instance that can be used to call Knora

        :param action: Action the object is used for (Action.CREATE or Action.UPDATE)
        :return: JSON-object
        """

        tmp = {}
        if action == Actions.Create:
            if self._username is None:
                raise BaseError("There must be a valid username!")
            tmp['username'] = self._username
            if self._email is None:
                raise BaseError("'email' is mandatory!")
            tmp['email'] = self._email
            if self._givenName is None:
                raise BaseError("'givenName is mandatory!")
            tmp['givenName'] = self._givenName
            if self._familyName is None:
                raise BaseError("'familyName' is mandatory!")
            tmp['familyName'] = self._familyName
            if self._password is None:
                raise BaseError("'password' is mandatory!")
            tmp['password'] = self._password
            if self._lang is None:
                raise BaseError("'language' is mandatory!")
            tmp['lang'] = self._lang.value
            tmp['status'] = True if self._status is None else self._status
            tmp['systemAdmin'] = False if self._sysadmin is None else self._sysadmin
        elif action == Actions.Update:
            tmp_changed = False
            if self._username is not None and 'username' in self._changed:
                tmp['username'] = self._username
                tmp_changed = self._username
            if self._email is not None and 'email' in self._changed:
                tmp['email'] = self._email
                tmp_changed = True
            if self._givenName is not None and 'givenName' in self._changed:
                tmp['givenName'] = self._givenName
                tmp_changed = True
            if self._familyName is not None and 'familyName' in self._changed:
                tmp['familyName'] = self._familyName
                tmp_changed = True
            if self._lang is not None and 'lang' in self._changed:
                tmp['lang'] = self._lang.value
                tmp_changed = True
            if not tmp_changed:
                tmp = {}
        return tmp

    def create(self) -> Any:
        """
        Create new user in Knora

        :return: JSON-object from Knora
        """

        jsonobj = self.toJsonObj(Actions.Create)
        jsondata = json.dumps(jsonobj)
        result = self._con.post('/admin/users', jsondata)
        id = result['user']['id']
        if self._in_projects is not None:
            for project in self._in_projects:
                result = self._con.post('/admin/users/iri/' + quote_plus(id) + '/project-memberships/' + quote_plus(project))
                if self._in_projects[project]:
                    result = self._con.post('/admin/users/iri/' + quote_plus(id) + '/project-admin-memberships/' + quote_plus(project))
        if self._in_groups is not None:
            for group in self._in_groups:
                result = self._con.post('/admin/users/iri/' + quote_plus(id) + '/group-memberships/' + quote_plus(group))
        return User.fromJsonObj(self._con, result['user'])

    def read(self) -> Any:
        """
        Read the user information from Knora. The User object must have a valid id or email!

        :return: JSON-object from Knora
        """
        if self._id is not None:
            result = self._con.get('/admin/users/iri/' + quote_plus(self._id))
        elif self._email is not None:
            result = self._con.get('/admin/users/email/' + quote_plus(self._email))
        elif self._username is not None:
            result = self._con.get('/admin/users/username/' + quote_plus(self._username))
        else:
            raise BaseError('Either user-id or email is required!')
        return User.fromJsonObj(self._con, result['user'])

    def update(self, requesterPassword: Optional[str] = None) -> Any:
        """
        Udate the user info in Knora with the modified data in this user instance

        :param requesterPassword: Old password if a user wants to change it's own password
        :return: JSON-object from Knora
        """

        jsonobj = self.toJsonObj(Actions.Update)
        if jsonobj:
            jsondata = json.dumps(jsonobj)
            result = self._con.put('/admin/users/iri/' + quote_plus(self.id) + '/BasicUserInformation', jsondata)
        if 'status' in self._changed:
            jsonobj = {'status': self._status}
            jsondata = json.dumps(jsonobj)
            result = self._con.put('/admin/users/iri/' + quote_plus(self.id) + '/Status', jsondata)
        if 'password' in self._changed:
            if requesterPassword is None:
                raise BaseError("Requester's password is missing!")
            jsonobj = {
                "requesterPassword": requesterPassword,
                "newPassword": self._password
            }
            jsondata = json.dumps(jsonobj)
            result = self._con.put('/admin/users/iri/' + quote_plus(self.id) + '/Password', jsondata)
        if 'sysadmin' in self._changed:
            jsonobj = {'systemAdmin': self._sysadmin}
            jsondata = json.dumps(jsonobj)
            result = self._con.put('/admin/users/iri/' + quote_plus(self.id) + '/SystemAdmin', jsondata)
        for p in self._add_to_project.items():
            result = self._con.post('/admin/users/iri/' + quote_plus(self._id) + '/project-memberships/' + quote_plus(p[0]))
            if p[1]:
                result = self._con.post('/admin/users/iri/' + quote_plus(self._id) + '/project-admin-memberships/' + quote_plus(p[0]))

        for p in self._rm_from_project:
            if self._in_projects.get(p) is not None and self._in_projects[p]:
                result = self._con.delete('/admin/users/iri/' + quote_plus(self._id) + '/project-admin-memberships/' + quote_plus(p))
            result = self._con.delete('/admin/users/iri/' + quote_plus(self._id) + '/project-memberships/' + quote_plus(p))

        for p in self._change_admin.items():
            if not p[0] in self._in_projects:
                raise BaseError('user must be member of project!')
            if p[1]:
                result = self._con.post('/admin/users/iri/' + quote_plus(self._id) + '/project-admin-memberships/' + quote_plus(p[0]))
            else:
                result = self._con.delete('/admin/users/iri/' + quote_plus(self._id) + '/project-admin-memberships/' + quote_plus(p[0]))

        for p in self._add_to_group:
            print('/admin/users/iri/' + quote_plus(self._id) + '/group-memberships/' + quote_plus(p))
            result = self._con.post('/admin/users/iri/' + quote_plus(self._id) + '/group-memberships/' + quote_plus(p))
        for p in self._rm_from_group:
            result = self._con.delete('/admin/users/iri/' + quote_plus(self._id) + '/group-memberships/' + quote_plus(p))
        user = User(con=self._con, id=self._id).read()
        return user

    def delete(self):
        """
        Delete the user in nore (NOT YET IMPLEMENTED)
        :return: None
        """
        result = self._con.delete('/admin/users/iri/' + quote_plus(self._id))
        return User.fromJsonObj(self._con, result['user'])

    @staticmethod
    def getAllUsers(con: Connection) -> List[Any]:
        """
        Get a list of all users (static method)

        :param con: Connection instance
        :return: List of users
        """

        result = con.get('/admin/users')
        if 'users' not in result:
            raise BaseError("Request got no users!")
        return list(map(lambda a: User.fromJsonObj(con, a), result['users']))

    def print(self) -> None:
        """
        Prin user info to stdout

        :return: None
        """

        print('User info:')
        print('  Id:          {}'.format(self._id))
        print('  Username:    {}'.format(self._username))
        print('  Family name: {}'.format(self._familyName))
        print('  Given name:  {}'.format(self._givenName))
        print('  Language:    {}'.format(self._lang.value))
        print('  Status:      {}'.format(self._status))
        print('  Sysadmin:    {}'.format(self._sysadmin))
        print('  In projects:')
        if self._in_projects is not None:
            for p in self._in_projects:
                print('    {} : project admin: {}'.format(p, self._in_projects[p]))
        print('   In groups:')
        if self._in_groups is not None:
            for g in self._in_groups:
                print('    {}'.format(g))



if __name__ == '__main__':
    con = Connection('http://0.0.0.0:3333')
    con.login('root@example.com', 'test')

    users = User.getAllUsers(con)
    for u in users:
        uu = u.read()
        uu.print()

    print('======================================')

    new_user = User(
        con=con,
        username="lrosenth",
        email="lukas.rosenthaler@gmail.com",
        givenName="Lukas",
        familyName="Rosenthaler",
        password="test",
        status=True,
        lang=Languages.DE,
        sysadmin=True,
        in_projects= {
            "http://rdfh.ch/projects/0001": True,
            "http://rdfh.ch/projects/yTerZGyxjZVqFMNNKXCDPF": False
        },
        in_groups={"http://rdfh.ch/groups/00FF/images-reviewer"}
    ).create()
    new_user.print()

    new_user.status = False
    new_user = new_user.update()
    new_user.print()

    new_user.status = True
    #new_user.givenName = '--Lukas--'
    new_user.familyName = '--Rosenthaler--'
    new_user.password = 'gaga'
    new_user = new_user.update("test")
    new_user.print()

    new_user.addToProject("http://rdfh.ch/projects/0803", True)
    new_user.rmFromProject('http://rdfh.ch/projects/0001')
    new_user.makeProjectAdmin('http://rdfh.ch/projects/yTerZGyxjZVqFMNNKXCDPF')
    new_user = new_user.update()
    new_user.print()



from typing import List, Set, Dict, Tuple, Optional, Any, Union
from urllib.parse import quote_plus
from rdflib import Graph
from lxml import etree
import requests
import json
import urllib
import pprint
import validators
import re
from rfc3987 import parse
from pprint import pprint
import sys

# TODO: recheck all the documentation of this file
"""
 Properties in knora-api:

 - :hasValue
 - :hasColor
 - :hasComment
 - :hasGeometry
 - :hasLinkTo
 - :isPartOf
 - :isRegionOf
 - :isAnnotationOf
 - :seqnum

 Classes in knora-api:
 - :Resource
 - :StillImageRepresentation
 - :TextRepresentation
 - :AudioRepresentation
 - :DDDRepresentation
 - :DocumentRepresentation
 - :MovingImageRepresentation
 - :Annotation -> :hasComment, :isAnnotationOf, :isAnnotationOfValue
 - :LinkObj -> :hasComment, :hasLinkTo, :hasLinkToValue
 - :LinkValue [reification node]
 - :Region -> :hasColor, :isRegionOf, :hasGeometry, :isRegionOfValue, :hasComment

 For lists:

 - :ListNode -> :hasSubListNode, :listNodePosition, :listNodeName, :isRootNode, :hasRootNode, :attachedToProject

 Values in knora-api:

 - :Value
 - :TextValue       -> :SimpleText, :TextArea
 - :ColorValue      -> :Colorpicker
 - :DateValue       -> :Date
 - :DecimalValue    -> :SimpleText
 - :GeomValue       -> :Geometry
 - :GeonameValue    -> :Geonames
 - :IntValue        -> :SimpleText, :Spinbox, :Slider
 - :BooleanValue    -> :Checkbox
 - :UriValue        -> :SimpleText
 - :IntervalValue
 - :ListValue       -> :Pulldown

 GUI elements

 - :Colorpicker     -> ncolors=integer
 - :Date
 - :Geometry
 - :Geonames
 - :Interval
 - :List            -> hlist(required)=<iri>
 - :Pulldown        -> hlist(required)=<iri>
 - :Radio           -> hlist(required)=<iri>
 - :Richtext
 - :Searchbox       -> numprops=integer
 - :SimpleText      -> maxlength=integer, size=integer
 - :Slider          -> max(required)=decimal, min(required)=decimal
 - :Spinbox         -> max=decimal, min=decimal
 - :Textarea        -> cols=integer, rows=integer, width=percent, wrap=string(soft|hard)
 - :Checkbox
 - :Fileupload
"""


class KnoraError(Exception):
    """Handles errors happening in this file"""

    def __init__(self, message):
        self.message = message


class KnoraStandoffXml:
    """Used to handle XML strings for standoff markup"""

    iriregexp = re.compile(r'IRI:[^:]*:IRI')

    def __init__(self, xmlstr: str):
        self.xmlstr = xmlstr

    def getXml(self):
        return self.xmlstr

    def findall(self):
        return KnoraStandoffXml.iriregexp.findall(self.xmlstr)

    def replace(self, fromStr: str, toStr: str):
        self.xmlstr.replace(fromStr, toStr)


class KnoraStandoffXmlEncoder(json.JSONEncoder):
    """Classes used as wrapper for knora standoff-XML"""
    def default(self, obj):
        if isinstance(obj, KnoraStandoffXml):
            return obj.getXml()
        return json.JSONEncoder.default(self, obj)


class Knora:
    """
    This is the main class which holds all the methods for communication with the Knora backend.
    """

    def __init__(self, server: str, prefixes: Dict[str, str] = None):
        """
        Constructor requiring the server address, the user and password of KNORA
        :param server: Address of the server, e.g https://api.dasch.swiss
        :param prefixes: Ontology prefixes used
        """
        self.server = server
        self.prefixes = prefixes
        self.token = None

    def login(self, email: str, password: str) -> None:
        """
        Method to login into KNORA which creates a session token.
        :param email: Email of user, e.g., root@example.com
        :param password: Password of the user, e.g. test
        """
        credentials = {
            "email": email,
            "password": password
        }
        jsondata = json.dumps(credentials)

        req = requests.post(
            self.server + '/v2/authentication',
            headers={'Content-Type': 'application/json; charset=UTF-8'},
            data=jsondata
        )
        self.on_api_error(req)
        result = req.json()
        self.token = result["token"]

    def get_token(self) -> str:
        return self.token

    def logout(self) -> None:
        if self.token is not None:
            req = requests.delete(
                self.server + '/v2/authentication',
                headers={'Authorization': 'Bearer ' + self.token}
            )
            self.on_api_error(req)
            self.token = None

    def __del__(self):
        self.logout()

    def on_api_error(self, res) -> None:
        """
        Method to check for any API errors
        :param res: The input to check, usually JSON format
        :return: Possible KnoraError that is being raised
        """

        if (res.status_code != 200):
            raise KnoraError("KNORA-ERROR: status code=" + str(res.status_code) + "\nMessage:" + res.text)

        if 'error' in res:
            raise KnoraError("KNORA-ERROR: API error: " + res.error)

    #==========================================================================
    # project related methods
    #

    def get_existing_projects(self, full: bool = False) -> List[Any]:
        """Returns a list of existing projects

        :return: List of existing projects
        """

        req = requests.get(self.server + '/admin/projects',
                           headers={'Authorization': 'Bearer ' + self.token})
        self.on_api_error(req)
        result = req.json()

        if 'projects' not in result:
            raise KnoraError("KNORA-ERROR:\n Request got no projects!")
        else:
            if full:
                return result['projects']
            else:
                return list(map(lambda a: a['id'], result['projects']))

    def get_project(self, shortcode: str) -> Dict[str,Any]:
        """Returns project data of given project

        :param shortcode: Shortcode of object
        :return: JSON containing the project information
        """

        url = self.server + '/admin/projects/shortcode/' + shortcode
        req = requests.get(url, headers={'Authorization': 'Bearer ' + self.token})
        self.on_api_error(req)

        result = req.json()

        return result["project"]

    def project_exists(self, proj_iri: str) -> bool:
        """Checks if a given project exists

        :return: Boolean
        """

        projects = self.get_existing_projects()
        return proj_iri in projects

    def create_project(
        self,
        shortcode: str,
        shortname: str,
        longname: str,
        descriptions: Optional[Dict[str, str]] = None,
        keywords: Optional[List[str]] = None,
        logo: Optional[str] = None) -> str:
        """
        Create a new project

        :param shortcode: Dedicated shortcode of project
        :param shortname: Short name of the project (e.g acronym)
        :param longname: Long name of project
        :param descriptions: Dict of the form {lang: descr, …} for the description of the project [Default: None]
        :param keywords: List of keywords
        :param logo: Link to the project logo [default: None]
        :return: Project IRI
        """

        descriptions = list(map(lambda p: {"language": p[0], "value": p[1]}, descriptions.items()))

        project = {
            "shortname": shortname,
            "shortcode": shortcode,
            "longname": longname,
            "status": True,
            "selfjoin": False
        }
        if descriptions is not None:
            project['description'] = descriptions
        if keywords is not None:
            project['keywords'] = keywords
        if logo is not None:
            project['logo'] = logo

        jsondata = json.dumps(project)
        # print(jsondata)

        req = requests.post(self.server + "/admin/projects",
                            headers={'Content-Type': 'application/json; charset=UTF-8',
                                     'Authorization': 'Bearer ' + self.token},
                            data=jsondata)
        self.on_api_error(req)

        res = req.json()
        return res["project"]["id"]

    def update_project(
        self,
        shortcode: str,
        shortname: Optional[str] = None,
        longname: Optional[str] = None,
        descriptions: Optional[Dict[str, str]] = None,
        keywords: Optional[List[str]] = None,
        logo: Optional[str] = None) -> str:
        """
        Update project information

        :param shortcode:
        :param shortname:
        :param longname:
        :param descriptions:
        :param keywords:
        :param logo:
        :return:
        """

        descriptions = list(map(lambda p: {"language": p[0], "value": p[1]}, descriptions.items()))

        project = {
            "longname": longname,
            "description": descriptions,
            "keywords": keywords,
            "logo": logo,
            "status": True,
            "selfjoin": False
        }

        jsondata = json.dumps(project)
        url = self.server + '/admin/projects/iri/' + quote_plus("http://rdfh.ch/projects/" + shortcode)

        req = requests.put(url,
                           headers={'Content-Type': 'application/json; charset=UTF-8',
                                    'Authorization': 'Bearer ' + self.token},
                           data=jsondata)
        self.on_api_error(req)

        res = req.json()
        return res['project']['id']

    #==========================================================================
    # Group related methods
    #

    def get_groups(self) -> List[Dict[str,Any]]:
        """
        Returns the list of existing groups

        :return: List of projects
        """
        url = self.server + '/admin/groups'

        req = requests.get(url, headers={'Authorization': 'Bearer ' + self.token})

        self.on_api_error(req)
        res = req.json()

        return res['groups']

    def get_group_by_iri(self, group_iri: str) -> Dict[str,Any]:
        """
        Returns information about the given group
        :param group_iri: IRI of the group
        :return: Information about the specific group
        """
        url = self.server + '/admin/groups/' + quote_plus(group_iri)

        req = requests.get(url, headers={'Authorization': 'Bearer ' + self.token})

        self.on_api_error(req)
        res = req.json()

        return res['group']

    def get_group_by_pshortname_and_gname(self,
                                          project_shortname: str,
                                          group_name: str) -> Union[str,None]:
        """
        Get a group by project shortname and group name

        :param project_shortname: Project shortname
        :param group_name: Group name
        :return: IRI of the group
        """
        groupinfos = self.get_groups()
        for groupinfo in groupinfos:
            if groupinfo["project"]["shortname"] == project_shortname and groupinfo["name"] == group_name:
                return groupinfo["id"]
        return None

    def get_group_by_pshortcode_and_gname(self,
                                          project_shortcode: str,
                                          group_name: str) -> Union[str, None]:
        """
        Get a group by project shortcode and group name

        :param project_shortname: Project shortcode
        :param group_name: Group name
        :return: IRI of the group
        """
        groupinfos = self.get_groups()
        for groupinfo in groupinfos:
            if groupinfo["project"]["shortcode"] == project_shortcode and groupinfo["name"] == group_name:
                return groupinfo["id"]
        return None

    def get_group_by_piri_and_gname(self,
                                    project_iri: str,
                                    group_name: str) -> Union[str, None]:
        """
        Get a group by project shortcode and group name

        :param project_shortname: Project shortcode
        :param group_name: Group name
        :return: IRI of the group
        """
        groupinfos = self.get_groups()
        for groupinfo in groupinfos:
            if groupinfo["project"]["id"] == project_iri and groupinfo["name"] == group_name:
                return groupinfo["id"]
        return None

    def create_group(self,
                     project_iri: str,
                     name: str,
                     description: Union[str, Dict[str,str]],
                     status: bool = True,
                     selfjoin: bool = False) -> str:
        """
        Create a new group

        :param name: Name of the group
        :param description: Either a string with the descrioption, or a List of Dicts in the form [{"value": "descr", "language": "lang"},…]
        :param project_iri: IRI of the project where the group belongs to
        :param status: Active (True) or not active (False) [default: True]
        :param selfjoin: ?? [default: False]
        :return: IRI of the group
        """

        groupinfo = {
            "name": name,
            "description": description if isinstance(description, str) else list(map(lambda p: {"@language": p[0], "@value": p[1]}, description.items())),
            "project": project_iri,
            "status": status,
            "selfjoin": selfjoin
        }
        jsondata = json.dumps(groupinfo)

        url = self.server + '/admin/groups'

        req = requests.post(url,
                            headers={'Content-Type': 'application/json; charset=UTF-8',
                                     'Authorization': 'Bearer ' + self.token},
                            data=jsondata)

        self.on_api_error(req)
        res = req.json()
        return res['group']['id']

    def update_group(self,
                     group_iri: str,
                     name: Optional[str] = None,
                     description: Optional[Union[str, Dict[str,str]]] = None,
                     selfjoin: Optional[bool] = None) -> Union[str,None]:
        """
        Modify the data about a group. Only parameters that have to be changed must be indicated
        :param group_iri: IRI of the grouo to be modified
        :param name: New name of the group [optional]
        :param description: Either a string with the descrioption, or a List of Dicts in the form [{"value": "descr", "language": "lang"},…] [optional]
        :param selfjoin: True or False [optional]
        :return: ???
        """

        groupinfo = {}
        done = False
        if name is not None:
            groupinfo['name'] = name
            done = True
        if description is not None:
            groupinfo['description'] = description if isinstance(description, str) else list(map(lambda p: {"@language": p[0], "@value": p[1]}, description.items()))
            done = True
        if selfjoin is not None:
            groupinfo['selfjoin'] = selfjoin
            done = True
        if done:
            jsondata = json.dumps(groupinfo)

            url = self.server + '/admin/groups/' + quote_plus(group_iri)

            req = requests.put(url,
                                headers={'Content-Type': 'application/json; charset=UTF-8',
                                         'Authorization': 'Bearer ' + self.token},
                                data=jsondata)
            self.on_api_error(req)
            res = req.json()
            pprint(res)
            return res['group']['id']
        else:
            return None

    def change_group_status(self,
                            group_iri: str,
                            status: bool) -> None:
        """
        Change the status of th group

        :param group_iri: IRI of the group
        :param status: Status (active: True, inactive: False)

        :return: None
        """

        statusinfo = {
            "status": status
        }
        jsondata = json.dumps(statusinfo)

        url = self.server + '/admin/groups/' + quote_plus(group_iri) + '/status'

        req = requests.put(url,
                           headers={'Content-Type': 'application/json; charset=UTF-8',
                                    'Authorization': 'Bearer ' + self.token},
                           data=jsondata)
        self.on_api_error(req)
        res = req.json()
        pprint(res)

    def delete_group(self,
                     group_iri: str) -> None:
        """
        Delete a group
        :param group_iri: IRI of the group
        :return:
        """
        url = self.server + '/admin/groups/' + quote_plus(group_iri)

        req = requests.delete(url,
                              headers={'Content-Type': 'application/json; charset=UTF-8',
                                       'Authorization': 'Bearer ' + self.token})
        self.on_api_error(req)
        res = req.json()
        pprint(res)

    #==========================================================================
    #  User related methods
    #

    def get_users(self) -> List[Dict[str,Any]]:
        """
        Get a list of all users

        :return: Json result.
        """
        url = self.server + '/admin/users'
        req = requests.get(url, headers={'Authorization': 'Bearer ' + self.token})

        self.on_api_error(req)
        res = req.json()
        return res['users']

    def get_user_by_iri(self, user_iri: str):
        """
        Get single user

        :return:
        """
        url = self.server + '/admin/users/iri/' + quote_plus(user_iri)
        req = requests.get(url, headers={'Authorization': 'Bearer ' + self.token})

        self.on_api_error(req)
        res = req.json()
        return res['user']

    def get_user_by_email(self, email: str):
        """
        Get a list of all users

        :return:
        """
        url = self.server + '/admin/users/email/' + quote_plus(email)
        req = requests.get(url, headers={'Authorization': 'Bearer ' + self.token})

        self.on_api_error(req)
        res = req.json()
        return res['user']

    def create_user(self,
                    username: str,
                    email: str,
                    given_name: str,
                    family_name: str,
                    password: str,
                    lang: str = "en",
                    sysadmin: bool = False):
        """
        Create a new user

        :param username: The username for login purposes (must be unique)
        :param email: The email address of the user
        :param given_name: The given name (surname, "Vorname", ...)
        :param family_name: The family name
        :param password: The password for the user
        :param lang: language code, either "en", "de", "fr", "it" [default: "en"]
        :param sysadmin: True if the user has system admin rights
        :return: The user ID as IRI
        """

        userinfo = {
            "username": username,
            "email": email,
            "givenName": given_name,
            "familyName": family_name,
            "password": password,
            "status": True,
            "lang": lang,
            "systemAdmin": sysadmin
        }

        jsondata = json.dumps(userinfo)
        url = self.server + '/admin/users'

        req = requests.post(url,
                            headers={'Content-Type': 'application/json; charset=UTF-8',
                                     'Authorization': 'Bearer ' + self.token},
                            data=jsondata)

        self.on_api_error(req)
        res = req.json()

        return res['user']['id']

    def update_user(self,
                    user_iri: str,
                    username: Optional[str] = None,
                    email:  Optional[str] = None,
                    given_name: Optional[str] = None,
                    family_name: Optional[str] = None,
                    password: Optional[str] = None,
                    lang: Optional[str] = None):
        userinfo: Dict[str,Any] = {};
        if username is not None:
            userinfo["username"] = username
        if email is not None:
            userinfo["email"] = email
        if given_name is not None:
            userinfo["givenName"] = given_name
        if family_name is not None:
            userinfo["familyName"] = family_name
        #if password is not None:
        #    update_user["password"] = password
        if lang is not None:
            userinfo["lang"] = lang
        if len(userinfo) > 0:
            url = self.server + '/admin/users/iri/' + quote_plus(user_iri) + '/BasicUserInformation'
            jsondata = json.dumps(userinfo)
            req = requests.put(url,
                               headers={'Content-Type': 'application/json; charset=UTF-8',
                                        'Authorization': 'Bearer ' + self.token},
                               data=jsondata)
            self.on_api_error(req)

    def change_user_password(self,
                             user_iri: str,
                             admin_password: str,
                             new_password: str):
        data = {
            "requesterPassword": admin_password,
            "newPassword": new_password
        }
        url = self.server + '/admin/users/iri/' + quote_plus(user_iri) + '/Password'
        jsondata = json.dumps(data)
        req = requests.put(url,
                           headers={'Content-Type': 'application/json; charset=UTF-8',
                                    'Authorization': 'Bearer ' + self.token},
                           data=jsondata)
        self.on_api_error(req)

    def add_user_to_project(self,
                            user_iri: str,
                            project_iri: str):
        """
        Add a user to a project

        :param user_iri: IRI of the user
        :param project_iri: IRI of the project
        :return: None
        """
        url = self.server + '/admin/users/iri/' + quote_plus(user_iri) + '/project-memberships/'\
              + quote_plus(project_iri)
        req = requests.post(url, headers={'Authorization': 'Bearer ' + self.token})
        self.on_api_error(req)

        return None

    def rm_user_from_project(self,
                            user_iri: str,
                            project_iri: str):
        """
        Remove a user from a project

        :param user_iri: IRI of the user
        :param project_iri: IRI of the project
        :return: None
        """
        url = self.server + '/admin/users/iri/' + quote_plus(user_iri) + '/project-memberships/' + quote_plus(
            project_iri)
        req = requests.delete(url, headers={'Authorization': 'Bearer ' + self.token})
        self.on_api_error(req)

        return None

    def add_user_to_project_admin(self,
                                  user_iri: str,
                                  project_iri: str) -> None:
        """
        Add a user to the project admin group (knora-admin:ProjectAdmin)

        :param user_iri: IRI of user
        :param project_iri: IRI of project
        :return: None
        """
        url = self.server + '/admin/users/iri/' + quote_plus(user_iri) + '/project-admin-memberships/' + quote_plus(
            project_iri)
        req = requests.post(url, headers={'Authorization': 'Bearer ' + self.token})
        self.on_api_error(req)
        return None

    def rm_user_from_project_admin(self,
                                  user_iri: str,
                                  project_iri: str) -> None:
        """
        Remove a user from the project admin group
        :param user_iri: IRI of user
        :param project_iri: IRI of project
        :return: None
        """
        url = self.server + '/admin/users/iri/' + quote_plus(user_iri) + '/project-admin-memberships/' + quote_plus(
            project_iri)
        req = requests.delete(url, headers={'Authorization': 'Bearer ' + self.token})
        self.on_api_error(req)
        return None

    def add_user_to_sysadmin(self, user_iri: str) -> None:
        """
        Add a user to the project admin group (knora-admin:ProjectAdmin)

        :param user_iri: IRI of user
        :param project_iri: IRI of project
        :return: None
        """
        data = {
            "systemAdmin": True
        }
        url = self.server + '/admin/users/iri/' + quote_plus(user_iri) + '/SystemAdmin'
        jsondata = json.dumps(data)
        req = requests.put(url, headers={'Content-Type': 'application/json; charset=UTF-8',
                                         'Authorization': 'Bearer ' + self.token},
                            data=jsondata)
        self.on_api_error(req)
        return None

    def rm_user_from_sysadmin(self, user_iri: str) -> None:
        """
        Remove a user from the system admin group

        :param user_iri: IRI of user
        :param project_iri: IRI of project
        :return: None
        """
        data = {
            "systemAdmin": False
        }
        url = self.server + '/admin/users/iri/' + quote_plus(user_iri) + '/SystemAdmin'
        jsondata = json.dumps(data)
        req = requests.put(url, headers={'Content-Type': 'application/json; charset=UTF-8',
                                         'Authorization': 'Bearer ' + self.token},
                              data=jsondata)
        self.on_api_error(req)
        return None

    def add_user_to_group(self,
                          user_iri: str,
                          group_iri: str) -> None:
        url = self.server + '/admin/users/iri/' + quote_plus(user_iri) + '/group-memberships/' + quote_plus(group_iri)

        req = requests.post(url, headers={'Authorization': 'Bearer ' + self.token})
        self.on_api_error(req)
        return None

    def rm_user_from_group(self,
                          user_iri: str,
                          group_iri: str) -> None:
        url = self.server + '/admin/users/iri/' + quote_plus(user_iri) + '/group-memberships/' + quote_plus(group_iri)

        req = requests.delete(url, headers={'Authorization': 'Bearer ' + self.token})
        self.on_api_error(req)
        return None

    #==========================================================================
    # Ontology methods
    #

    def get_existing_ontologies(self):
        """

        :return: Returns the metadata of all existing ontologies on v2/ontologies
        """

        req = requests.get(self.server + '/v2/ontologies/metadata',
                           headers={'Authorization': 'Bearer ' + self.token})
        result = req.json()

        if not '@graph' in result:
            raise KnoraError("KNORA-ERROR:\n Request got no graph!")
        else:
            names = list(map(lambda a: a['@id'], result['@graph']))
            return names

    def get_project_ontologies(self, project_code: str) -> Optional[dict]:
        """

        :param project_code:
        :return:
        """

        proj = quote_plus("http://rdfh.ch/projects/" + project_code)
        req = requests.get(self.server + "/v2/ontologies/metadata/" + proj,
                           headers={'Authorization': 'Bearer ' + self.token})
        self.on_api_error(req)
        result = req.json()

        if '@graph' in result:  # multiple ontologies
            ontos = list(map(lambda a: {
                'iri': a['@id'],
                'label': a['rdfs:label'],
                'moddate': a.get('knora-api:lastModificationDate')
            }, result['@graph']))
            return ontos
        elif '@id' in result:  # single ontology
            return [{
                'iri': result['@id'],
                'label': result['rdfs:label'],
                'moddate': result.get('knora-api:lastModificationDate')
            }]
        else:
            return None

    def ontology_exists(self, onto_iri: str):
        """
        Checks if an ontology exists

        :param onto_iri: The possible ontology iri
        :return: boolean
        """

        ontos = self.get_existing_ontologies()

        return onto_iri in ontos

    def get_ontology_lastmoddate(self, onto_iri: str):
        """
        Retrieves the lastModificationDate of a Ontology

        :param onto_iri: The ontology to retrieve the lastModificationDate from.
        :return: The lastModificationDate if it exists. Else, this method returns a dict with (id, None). If the ontology does not exist, it return None.
        """

        req = requests.get(self.server + '/v2/ontologies/metadata',
                           headers={'Authorization': 'Bearer ' + self.token})
        result = req.json()

        all_ontos = {}

        for onto in result['@graph']:
            if 'knora-api:lastModificationDate' in onto:
                all_ontos.__setitem__(onto['@id'], onto['knora-api:lastModificationDate'])
            else:
                all_ontos.__setitem__(onto['@id'], None)

        return all_ontos[onto_iri]

    def create_ontology(self,
                        onto_name: str,
                        project_iri: str,
                        label: str) -> Dict[str, str]:
        """
        Create a new ontology

        :param onto_name: Name of the omntology
        :param project_iri: IRI of the project
        :param label: A label property for this ontology
        :return: Dict with "onto_iri" and "last_onto_date"
        """

        ontology = {
            "knora-api:ontologyName": onto_name,
            "knora-api:attachedToProject": {
                "@id": project_iri
            },
            "rdfs:label": label,
            "@context": {
                "rdfs": 'http://www.w3.org/2000/01/rdf-schema#',
                "knora-api": 'http://api.knora.org/ontology/knora-api/v2#'
            }
        }

        jsondata = json.dumps(ontology)

        req = requests.post(self.server + "/v2/ontologies",
                            headers={'Content-Type': 'application/json; charset=UTF-8',
                                     'Authorization': 'Bearer ' + self.token},
                            data=jsondata)

        self.on_api_error(req)

        res = req.json()
        # TODO: return also ontology name
        return {"onto_iri": res['@id'], "last_onto_date": res['knora-api:lastModificationDate']}

    def delete_ontology(self, onto_iri: str, last_onto_date=None):
        """
        A method to delete an ontology from /v2/ontologies

        :param onto_iri: The ontology to delete
        :param last_onto_date: the lastModificationDate of an ontology. None by default
        :return:
        """""  # TODO: add return documentation
        url = self.server + "/v2/ontologies/" + urllib.parse.quote_plus(onto_iri)
        req = requests.delete(url,
                              params={"lastModificationDate": last_onto_date},
                              headers={'Authorization': 'Bearer ' + self.token})
        self.on_api_error(req)
        res = req.json()
        return res

    def get_ontology_graph(self,
                           shortcode: str,
                           name: str):
        """
        Returns the turtle definition of the ontology.

        :param shortcode: Shortcode of the project
        :param name: Name of the ontology
        :return:
        """
        url = self.server + "/ontology/" + shortcode + "/" + name + "/v2"
        turtle = requests.get(url,
                              headers={"Accept": "text/turtle",
                                       'Authorization': 'Bearer ' + self.token})
        self.on_api_error(turtle)
        return turtle.text

    def create_res_class(self,
                         onto_iri: str,
                         onto_name: str,
                         last_onto_date: str,
                         class_name: str,
                         super_class: List[str],
                         labels: Dict[str, str],
                         comments: Optional[Dict[str, str]] = None,
                         permissions: Optional[Dict[str, str]] = None) -> Dict[str, str]:
        """Creates a knora resource class

        :param onto_iri: IRI of the ontology
        :param onto_name: Name of the ontology
        :param last_onto_date: Last modification date as returned by last call
        :param class_name: Name of the class to be created
        :param super_class: List of super classes
        :param labels: Dict with labels in the form { lang: labeltext }
        :param comments: Dict with comments in the form { lang: commenttext }
        :param permissions: Dict with permissions in the form
        :return: Dict with "class_iri" and "last_onto_date"
        """

        #
        # using map and iterable to get the proper format
        #
        labels = list(map(lambda p: {"@language": p[0], "@value": p[1]}, labels.items()))

        if not comments:
            comments = {"en": "none"}

        #
        # using map and iterable to get the proper format
        #
        comments = list(map(lambda p: {"@language": p[0], "@value": p[1]}, comments.items()))

        res_class = {
            "@id": onto_iri,
            "@type": "owl:Ontology",
            "knora-api:lastModificationDate": last_onto_date,
            "@graph": [{
                "@id": onto_name + ":" + class_name,
                "@type": "owl:Class",
                "rdfs:label": labels,
                "rdfs:comment": comments,
                "rdfs:subClassOf": {
                    "@id": super_class
                }
            }],
            "@context": {
                "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
                "knora-api": "http://api.knora.org/ontology/knora-api/v2#",
                "owl": "http://www.w3.org/2002/07/owl#",
                "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                "xsd": "http://www.w3.org/2001/XMLSchema#",
                onto_name: onto_iri + "#"
            }
        }

        jsondata = json.dumps(res_class, indent=3, separators=(',', ': '))

        req = requests.post(self.server + "/v2/ontologies/classes",
                            headers={'Content-Type': 'application/json; charset=UTF-8',
                                     'Authorization': 'Bearer ' + self.token},
                            data=jsondata)
        self.on_api_error(req)

        res = req.json()
        return {"class_iri": res['@graph'][0]['@id'], "last_onto_date": res['knora-api:lastModificationDate']}

    def create_property(
        self,
        onto_iri: str,
        onto_name: str,
        last_onto_date: str,
        prop_name: str,
        super_props: List[str],
        labels: Dict[str, str],
        gui_element: str,
        gui_attributes: List[str] = None,
        subject: Optional[str] = None,
        object: Optional[str] = None,
        comments: Optional[Dict[str, str]] = None
    ) -> Dict[str, str]:
        """Create a Knora property

        :param onto_iri: IRI of the ontology
        :param onto_name: Name of the Ontology (prefix)
        :param last_onto_date: Last modification date as returned by last call
        :param prop_name: Name of the property
        :param super_props: List of super-properties
        :param labels: Dict with labels in the form { lang: labeltext }
        :param gui_element: Valid GUI-Element
        :param gui_attributes: Valid GUI-Attributes (or None)
        :param subject: Full name (prefix:name) of subject resource class
        :param object: Full name (prefix:name) of object resource class
        :param comments: Dict with comments in the form { lang: commenttext }
        :return: Dict with "prop_iri" and "last_onto_date" keys
        """
        #
        # using map and iterable to get the proper format
        #
        labels = list(map(lambda p: {"@language": p[0], "@value": p[1]}, labels.items()))

        if not comments:
            comments = {"en": "none"}

        #
        # using map and iterable to get the proper format
        #
        comments = list(map(lambda p: {"@language": p[0], "@value": p[1]}, comments.items()))

        additional_context = {}
        for sprop in super_props:
            pp = sprop.split(':')
            if pp[0] != "knora-api":
                additional_context[pp[0]] = self.prefixes[pp[0]]

        #
        # using map and iterable to get the proper format
        #
        super_props = list(map(lambda x: {"@id": x}, super_props))
        if len(super_props) == 1:
            super_props = super_props[0]

        propdata = {
            "@id": onto_name + ":" + prop_name,
            "@type": "owl:ObjectProperty",
            "rdfs:label": labels,
            "rdfs:comment": comments,
            "rdfs:subPropertyOf": super_props,
            "salsah-gui:guiElement": {
                "@id": gui_element
            }
        }
        if subject:
            propdata["knora-api:subjectType"] = {
                "@id": subject
            }

        if object:
            propdata["knora-api:objectType"] = {
                "@id": object
            }

        if gui_attributes:
            propdata["salsah-gui:guiAttribute"] = gui_attributes

        property = {
            "@id": onto_iri,
            "@type": "owl:Ontology",
            "knora-api:lastModificationDate": last_onto_date,
            "@graph": [
                propdata
            ],
            "@context": {
                "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
                "knora-api": "http://api.knora.org/ontology/knora-api/v2#",
                "salsah-gui": "http://api.knora.org/ontology/salsah-gui/v2#",
                "owl": "http://www.w3.org/2002/07/owl#",
                "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                "xsd": "http://www.w3.org/2001/XMLSchema#",
                onto_name: onto_iri + "#"
            }
        }
        property["@context"].update(additional_context)
        jsondata = json.dumps(property, indent=3, separators=(',', ': '))
        req = requests.post(self.server + "/v2/ontologies/properties",
                            headers={'Content-Type': 'application/json; charset=UTF-8',
                                     'Authorization': 'Bearer ' + self.token},
                            data=jsondata)
        self.on_api_error(req)

        res = req.json()
        return {"prop_iri": res['@graph'][0]['@id'], "last_onto_date": res['knora-api:lastModificationDate']}

    def create_cardinality(
        self,
        onto_iri: str,
        onto_name: str,
        last_onto_date: str,
        class_iri: str,
        prop_iri: str,
        occurrence: str,
        gui_order: Optional[int] = None
    ) -> Dict[str, str]:
        """Add a property with a given cardinality to a class

        :param onto_iri: IRI of the ontology
        :param onto_name: Name of the ontology (prefix)
        :param last_onto_date: Last modification date as returned by last call
        :param class_iri: IRI of the class to which the property will be added
        :param prop_iri: IRI of the property that should be added
        :param occurrence: Occurrence: "1", "0-1", "0-n" or "1-n"
        :param gui_order: Ordering of properties in GUI
        :return: Dict with "last_onto_date" key
        """
        switcher = {
            "1": ("owl:cardinality", 1),
            "0-1": ("owl:maxCardinality", 1),
            "0-n": ("owl:minCardinality", 0),
            "1-n": ("owl:minCardinality", 1)
        }
        occurrence = switcher.get(occurrence)
        if not occurrence:
            KnoraError("KNORA-ERROR:\n Invalid occurrence!")

        cardinality = {
            "@id": onto_iri,
            "@type": "owl:Ontology",
            "knora-api:lastModificationDate": last_onto_date,
            "@graph": [{
                "@id": class_iri,
                "@type": "owl:Class",
                "rdfs:subClassOf": {
                    "@type": "owl:Restriction",
                    occurrence[0]: occurrence[1],
                    "owl:onProperty": {
                        "@id": prop_iri
                    }
                }
            }],
            "@context": {
                "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
                "owl": "http://www.w3.org/2002/07/owl#",
                "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                "xsd": "http://www.w3.org/2001/XMLSchema#",
                "knora-api": "http://api.knora.org/ontology/knora-api/v2#",
                "salsah-gui": "http://api.knora.org/ontology/salsah-gui/v2#",
                onto_name: onto_iri + "#"
            }
        }
        if gui_order is not None:
            cardinality['@graph'][0]["rdfs:subClassOf"]["salsah-gui:guiOrder"] = int(gui_order)




        jsondata = json.dumps(cardinality, indent=3, separators=(',', ': '))

        req = requests.post(self.server + "/v2/ontologies/cardinalities",
                            headers={'Content-Type': 'application/ld+json; charset=UTF-8',
                                     'Authorization': 'Bearer ' + self.token},
                            data=jsondata)
        self.on_api_error(req)

        res = req.json()

        return {"last_onto_date": res["knora-api:lastModificationDate"]}

    def create_list_node(self,
                         project_iri: str,
                         labels: Dict[str, str],
                         comments: Optional[Dict[str, str]] = None,
                         name: Optional[str] = None,
                         parent_iri: Optional[str] = None) -> str:
        """
        Creates a new list node. If there is no parent, a root node is created

        :param project_iri: IRI of the project
        :param labels: Dict in the form {lang: label, …} giving the label(s)
        :param comments: Dict in the form {lang: comment, …} giving the comment(s)
        :param name: Name of the list node
        :param parent_iri: None for root node (or omit), otherwise IRI of parent node
        :return: IRI of list node
        """

        #
        # using map and iterable to get the proper format
        #
        labels = list(map(lambda p: {"language": p[0], "value": p[1]}, labels.items()))

        listnode = {
            "projectIri": project_iri,
            "labels": labels,
        }

        #
        # using map and iterable to get the proper format
        #
        if comments is not None:
            listnode["comments"] = list(map(lambda p: {"language": p[0], "value": p[1]}, comments.items()))
        else:
            listnode["comments"] = []

        if name is not None:
            listnode["name"] = name

        if parent_iri is not None:
            listnode["parentNodeIri"] = parent_iri
            url = self.server + "/admin/lists/" + quote_plus(parent_iri)
        else:
            url = self.server + "/admin/lists"

        jsondata = json.dumps(listnode, indent=3, separators=(',', ': '))

        req = requests.post(url,
                            headers={'Content-Type': 'application/json; charset=UTF-8',
                                     'Authorization': 'Bearer ' + self.token},
                            data=jsondata)
        self.on_api_error(req)

        res = req.json()

        if parent_iri is not None:
            return res['nodeinfo']['id']
        else:
            return res['list']['listinfo']['id']

    def get_lists(self, shortcode: str):
        """
        Get the lists belonging to a certain project identified by its shortcode
        :param shortcode: Project shortcode

        :return: JSON with the lists
        """
        url = self.server + "/admin/lists?projectIri=" + quote_plus("http://rdfh.ch/projects/" + shortcode)
        req = requests.get(url, headers={'Authorization': 'Bearer ' + self.token})
        self.on_api_error(req)
        return req.json()

    def get_complete_list(self, list_iri: str):
        """
        Get all the data (nodes) of a specific list

        :param list_iri: IRI of the list
        :return: JSON containing the list info including all nodes
        """
        url = self.server + "/admin/lists/" + quote_plus(list_iri)
        req = requests.get(url, headers={'Authorization': 'Bearer ' + self.token})
        self.on_api_error(req)
        return req.json()

    def get_resource_by_label(self,
                              label: str,
                              res_class: Optional[str] = None,
                              limit_to_project: Optional[str] = None,
                              offset: Optional[int] = None):
        url = self.server + "/v2/searchbylabel/" + label
        option = False
        if res_class is not None:
            url += '?limitToResourceClass=' + quote_plus(res_class)
            option = True
        if limit_to_project is not None:
            if option:
                url += '&limitToProject=' + quote_plus(limit_to_project)
            else:
                url += '?limitToProject=' + quote_plus(limit_to_project)
            option = True;
        if offset is not None:
            if option:
                url += '&offset=' + quote_plus(limit_to_project)
            else:
                url += '&offset=' + quote_plus(limit_to_project)
        req = requests.get(url, headers={'Authorization': 'Bearer ' + self.token})
        self.on_api_error(req)
        return req.json()

    def create_resource(self,
                        schema: Dict,
                        res_class: str,
                        label: str,
                        values: Dict,
                        permissions: Optional[str] = None,
                        stillimage: Optional[str] = None):
        """
        This method creates a new resource (instance of a resource class) with the
        default permissions.

        :param schema: The schema of the ontology as returned by the method "create_schema()"
        :param res_class: The resource class of the resource to be created
        :param label: The "rdfs:label" to be given to the new resource
        :param values: A dict with the property values. It has the form
        { property_name: value, property_name: value,… } or { property_name: [value1, value2,…],… }
        The format of the values depends on the value types. E.g. a calendar date has the form
        "GREGORIAN:CE:1920-03-12:CE:1921:05:21" where all values except the start year are optional.
        :param stillimage: Path to a still image...
        :return: A dict in the form { 'iri': resource_iri, 'ark': ark_id, 'vark': dated_ark_id }
        """

        ontoname = schema["ontoname"]
        props = schema['resources'][res_class]  # this is an array of all properties defined in the ontology

        # we start building the dict that will be transformed into the JSON-LD
        jsondata = {
            '@type': ontoname + ":" + res_class,
            'rdfs:label': label,
            "knora-api:attachedToProject": {
                "@id": schema['proj_iri']
            }
        }

        if permissions is not None:
            jsondata["knora-api:hasPermissions"] = permissions

        if stillimage is not None:
            jsondata["knora-api:hasStillImageFileValue"] = {
                "@type": "knora-api:StillImageFileValue",
                "knora-api:fileValueHasFilename": stillimage
            }

        def create_valdict(val):
            """
            Internal function to create the JSON-LD for one value
            :param val: the value
            :return: Dict propared for the JSON-LD for one value

            """

            if type(val) is dict:
                pprint(val)
                comment = val.get('comment')
                permissions = val.get('permissions')
                mapping = val.get('mapping')
                val = val.get('value')
            else:
                comment = None
                permissions = None
                mapping = None

            valdict = {
                '@type': 'knora-api:' + prop["otype"]
            }

            if permissions is not None:
                valdict["knora-api:hasPermissions"] = permissions

            if comment is not None:
                valdict["knora-api:valueHasComment"] = comment

            if prop["otype"] == "TextValue":
                if isinstance(val, KnoraStandoffXml):  # text with XML markup
                    valdict['knora-api:textValueAsXml'] = val  # no conversion to string
                    valdict['knora-api:textValueHasMapping'] = {
                        '@id': 'http://rdfh.ch/standoff/mappings/StandardMapping' if mapping is None else mapping
                    }
                else:  # normal text string without markup
                    valdict['knora-api:valueAsString'] = str(val)
            elif prop["otype"] == "ColorValue":
                #
                # a color value as used in HTML (e.g. "#aaccff"
                #
                res = re.match('^#(?:[0-9a-fA-F]{3}){1,2}$', str(val))
                if res is None:
                    raise KnoraError("Invalid ColorValue format! " + str(val))
                valdict['knora-api:colorValueAsColor'] = str(val)
            elif prop["otype"] == "DateValue":
                #
                # A knora date value
                #
                res = re.match(
                    '(GREGORIAN:|JULIAN:)?(CE:|BCE:)?(\d{4})?(-\d{1,2})?(-\d{1,2})?(:CE|:BCE)?(:\d{4})?(-\d{1,2})?(-\d{1,2})?',
                    str(val))
                if res is None:
                    raise KnoraError("Invalid date format! " + str(val))
                dp = res.groups()
                calendar = 'GREGORIAN' if dp[0] is None else dp[0].strip('-: ')
                e1 = 'CE' if dp[1] is None else dp[1].strip('-: ')
                y1 = None if dp[2] is None else int(dp[2].strip('-: '))
                m1 = None if dp[3] is None else int(dp[3].strip('-: '))
                d1 = None if dp[4] is None else int(dp[4].strip('-: '))
                e2 = 'CE' if dp[5] is None else dp[5].strip('-: ')
                y2 = None if dp[6] is None else int(dp[6].strip('-: '))
                m2 = None if dp[7] is None else int(dp[7].strip('-: '))
                d2 = None if dp[8] is None else int(dp[8].strip('-: '))
                if y1 is None:
                    raise KnoraError("Invalid date format! " + str(val))
                if y2 is not None:
                    date1 = y1 * 10000
                    if m1 is not None:
                        date1 += m1 * 100
                    if d1 is not None:
                        date1 += d1
                    date2 = y2 * 10000
                    if m2 is not None:
                        date2 += m2 * 100
                    if d2 is not None:
                        date2 += d2
                    if date1 > date2:
                        y1, y2 = y2, y1
                        m1, m2 = m2, m1
                        d1, d2 = d2, d1
                valdict["knora-api:dateValueHasCalendar"] = calendar
                valdict["knora-api:dateValueHasStartEra"] = e1
                valdict["knora-api:dateValueHasStartYear"] = int(y1)
                if m1 is not None:
                    valdict["knora-api:dateValueHasStartMonth"] = int(m1)
                if d1 is not None:
                    valdict["knora-api:dateValueHasStartDay"] = int(d1)
                valdict["knora-api:dateValueHasEndEra"] = e2
                if y2 is not None:
                    valdict["knora-api:dateValueHasEndYear"] = int(y2)
                else:
                    valdict["knora-api:dateValueHasEndYear"] = int(y1)
                if m2 is not None:
                    valdict["knora-api:dateValueHasEndMonth"] = int(m2)
                if d2 is not None:
                    valdict["knora-api:dateValueHasEndDay"] = int(d2)
            elif prop["otype"] == "DecimalValue":
                #
                # a decimal value
                #
                valdict['knora-api:decimalValueAsDecimal'] = {
                    '@type': 'xsd:decimal',
                    '@value': str(val)
                }
            elif prop["otype"] == "GeomValue":
                #
                # A geometry ID
                #
                valdict['knora-api:geometryValueAsGeometry'] = str(val)
            elif prop["otype"] == "GeonameValue":
                #
                # A geoname ID
                #
                valdict['knora-api:geonameValueAsGeonameCode'] = str(val)
            elif prop["otype"] == "IntValue":
                #
                # an integer value
                #
                valdict['knora-api:intValueAsInt'] = int(val)
            elif prop["otype"] == "BooleanValue":
                #
                # a boolean value
                #
                if type(val) == bool:
                    valdict['knora-api:booleanValueAsBoolean'] = val
                elif type(val) == str:
                    if val.upper() == 'TRUE':
                        valdict['knora-api:booleanValueAsBoolean'] = True
                    elif val.upper() == 'FALSE':
                        valdict['knora-api:booleanValueAsBoolean'] = False
                    else:
                        raise KnoraError("Invalid boolean format! " + str(val))
                elif type(val) == int:
                    if val == 0:
                        valdict['knora-api:booleanValueAsBoolean'] = False
                    else:
                        valdict['knora-api:booleanValueAsBoolean'] = True
            elif prop["otype"] == "UriValue":
                #
                # an URI
                #
                valdict['knora-api:uriValueAsUri'] = {
                    "@type": "xsd:anyURI",
                    "@value": str(val)
                }
            elif prop["otype"] == "TimeValue":
                #
                # an URI
                #
                valdict['knora-api:timeValueAsTime'] = {
                    "@type": "xsd:dateTime",
                    "@value": str(val)
                }
            elif prop["otype"] == "IntervalValue":
                #
                # an interval in the form "1.356:2.456"
                #
                iv = val.split(':')
                valdict["knora-api:intervalValueHasEnd"] = {
                    "@type": "xsd:decimal",
                    "@value": str(iv[0])
                }
                valdict["knora-api:intervalValueHasStart"] = {
                    "@type": "xsd:decimal",
                    "@value": str(iv[1])
                }
            elif prop["otype"] == "ListValue":
                try:
                    iriparts = parse(str(val), rule='IRI')
                    if iriparts['scheme'] == 'http' or iriparts['scheme'] == 'https':
                        valdict['knora-api:listValueAsListNode'] = {
                            '@id': str(val)
                        }
                    else:
                        if iriparts['authority'] is not None:
                            raise KnoraError("Invalid list node: \"" + str(val) + "\" !")
                        listname = iriparts['scheme']
                        nodename = iriparts['path']
                        for node in schema['lists'][listname]['nodes']:
                            found = False
                            if node['name'] == nodename:
                                valdict['knora-api:listValueAsListNode'] = {
                                    '@id': node['id']
                                }
                                found = True
                                break
                        if not found:
                            raise KnoraError("Invalid list node: \"" + str(val) + "\" !")
                except ValueError as err:
                    raise KnoraError("Invalid list node: \"" + str(val) + "\" !")

            elif prop["otype"] == "LinkValue":
                valdict['@type'] = 'knora-api:LinkValue'
                valdict['knora-api:linkValueHasTargetIri'] = {
                    '@id': str(val)
                }
            else:
                if prop['otype'] in schema['link_otypes']:
                    valdict['@type'] = 'knora-api:LinkValue'
                    valdict['knora-api:linkValueHasTargetIri'] = {
                        '@id': str(val)
                    }
                else:
                    raise KnoraError("Invalid otype: " + prop['otype'])

            return valdict

        for key, value in values.items():
            prop = None
            for tmpprop in props:
                if tmpprop['propname'] == key:
                    prop = tmpprop
            if prop is None:
                raise KnoraError("Property " + key + " not known!")

            if prop['otype'] == "LinkValue" or prop['otype'] in schema['link_otypes']:
                nkey = key + "Value"
            else:
                nkey = key
            if type(value) is list:
                valarr = []
                for val in value:
                    valarr.append(create_valdict(val))
                jsondata[ontoname + ':' + nkey] = valarr
            else:
                jsondata[ontoname + ':' + nkey] = create_valdict(value)

            jsondata['@context'] = {
                "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
                "knora-api": "http://api.knora.org/ontology/knora-api/v2#",
                "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                "xsd": "http://www.w3.org/2001/XMLSchema#",
                ontoname: schema['onto_iri'] + '#'
            }

        jsonstr = json.dumps(jsondata, indent=3, separators=(',', ': '), cls=KnoraStandoffXmlEncoder)
        print(jsonstr)
        url = self.server + "/v2/resources"
        req = requests.post(url,
                            headers={'Content-Type': 'application/json; charset=UTF-8',
                                     'Authorization': 'Bearer ' + self.token},
                            data=jsonstr)
        self.on_api_error(req)

        res = req.json()

        return {
            'iri': res['@id'],
            'ark': res['knora-api:arkUrl']['@value'],
            'vark': res['knora-api:versionArkUrl']['@value']
        }

    def list_creator(self, children: List):
        """
        internal Helper function

        :param children:
        :return:
        """
        if len(children) == 0:
            res = list(map(lambda a: {"name": a["name"], "id": a["id"]}, children))
        else:
            res = list(
                map(lambda a: {"name": a["name"], "id": a["id"], "nodes": self.list_creator(a["children"])}, children))
        return res

    def create_schema(self, shortcode: str, shortname: str):
        """
        This method extracts the ontology from the ontology information it gets from Knora. It
        gets the ontology information as n3-data using the Knora API and concerts into a convenient
        python dict that can be used for further processing. It is required by the bulk import processing
        routines.

        :param shortcode: Shortcode of the project
        :param shortname: Short name of the ontolopgy
        :return: Dict with a simple description of the ontology
        """
        turtle = self.get_ontology_graph(shortcode, shortname)
        # print(turtle)
        g = Graph()
        g.parse(format='n3', data=turtle)

        # Get project and ontology IRI's
        sparql = """
        SELECT ?onto ?proj
        WHERE {
            ?onto a owl:Ontology .
            ?onto knora-api:attachedToProject ?proj
        }
        """
        qres = g.query(sparql)
        for row in qres:
            proj_iri = row.proj.toPython()  # project IRI
            onto_iri = row.onto.toPython()  # ontology IRI

        sparql = """
        SELECT ?res ?prop ?superprop ?otype ?guiele ?attr ?card ?cardval
        WHERE {
            ?res a owl:Class .
            ?res rdfs:subClassOf ?restriction .
            ?restriction a owl:Restriction .
            ?restriction owl:onProperty ?prop .
            ?restriction ?card ?cardval .
            ?prop a owl:ObjectProperty .
            ?prop knora-api:objectType ?otype .
            ?prop salsah-gui:guiElement ?guiele .
            ?prop rdfs:subPropertyOf ?superprop .
            OPTIONAL { ?prop salsah-gui:guiAttribute ?attr } .
            FILTER(?card = owl:cardinality || ?card = owl:maxCardinality || ?card = owl:minCardinality)
        }
        ORDER BY ?res ?prop
        """
        qres = g.query(sparql)

        resources = {}
        resclass = ''
        propname = ''
        link_otypes = []
        propcnt = 0
        propindex = {}  # we have to keep the order of the properties as given in the ontology....
        for row in qres:

            nresclass = row.res.toPython()
            nresclass = nresclass[nresclass.find('#') + 1:]
            if resclass != nresclass:
                resclass = nresclass
                resources[resclass] = []
                propcnt = 0
            superprop = row.superprop.toPython()
            superprop = superprop[superprop.find('#') + 1:]
            if superprop == 'hasLinkToValue':  # we ignore this one....
                continue
            npropname = row.prop.toPython()
            npropname = npropname[npropname.find('#') + 1:]
            attr = row.attr.toPython() if row.attr is not None else None
            if attr is not None:
                attr = attr.split('=')
            if propname == npropname:
                propcnt -= 1

                # process attribute (there might be multiple attributes)
                if attr is not None:
                    if resources[resclass][propcnt]["attr"] is not None:  # TODO: why is this necessary???
                        resources[resclass][propcnt]["attr"][attr[0]] = attr[1].strip('<>')

                # process superprop (there might be multiple superprops)
                if superprop not in resources[resclass][propcnt]["superprop"]:
                    resources[resclass][propcnt]["superprop"].append(superprop)
                # pprint.pprint(resources[resclass])
                propcnt += 1
                continue
            else:
                propname = npropname
            objtype = row.otype.toPython()
            objtype = objtype[objtype.find('#') + 1:]
            card = row.card.toPython()
            card = card[card.find('#') + 1:]
            guiele = row.guiele.toPython()
            guiele = guiele[guiele.find('#') + 1:]
            resources[resclass].append({
                "propname": propname,
                "otype": objtype,
                "superprop": [superprop],
                "guiele": guiele,
                "attr": {attr[0]: attr[1].strip('<>')} if attr is not None else None,
                "card": card,
                "cardval": row.cardval.toPython()
            })
            # pprint.pprint(resources[resclass])
            if superprop == "hasLinkTo":
                link_otypes.append(objtype)
            propindex[propname] = propcnt
            propcnt += 1

        # Get info about lists attached to the project
        listdata = {}
        lists = self.get_lists(shortcode)
        lists = lists["lists"]
        for list in lists:
            tmp = self.get_complete_list(list["id"])
            clist = tmp["list"]
            listdata[clist["listinfo"]["name"]] = {
                "id": clist["listinfo"]["id"],
                "nodes": self.list_creator(clist["children"])
            }
        schema = {
            "proj_iri": proj_iri,
            "shortcode": shortcode,
            "ontoname": shortname,
            "onto_iri": onto_iri,
            "lists": listdata,
            "resources": resources,
            "link_otypes": link_otypes
        }
        return schema

    def reset_triplestore_content(self):
        rdfdata = [
            {
                "path": "./knora-ontologies/knora-admin.ttl",
                "name": "http://www.knora.org/ontology/knora-admin"
            },
            {
                "path": "./knora-ontologies/knora-base.ttl",
                "name": "http://www.knora.org/ontology/knora-base"
            },
            {
                "path": "./knora-ontologies/standoff-onto.ttl",
                "name": "http://www.knora.org/ontology/standoff"
            },
            {
                "path": "./knora-ontologies/standoff-data.ttl",
                "name": "http://www.knora.org/data/standoff"
            },
            {
                "path": "./knora-ontologies/salsah-gui.ttl",
                "name": "http://www.knora.org/ontology/salsah-gui"
            },
            {
                "path": "./_test_data/all_data/admin-data.ttl",
                "name": "http://www.knora.org/data/admin"
            },
            {
                "path": "./_test_data/all_data/permissions-data.ttl",
                "name": "http://www.knora.org/data/permissions"
            },
            {
                "path": "./_test_data/all_data/system-data.ttl",
                "name": "http://www.knora.org/data/0000/SystemProject"
            }
        ]
        jsondata = json.dumps(rdfdata)
        url = self.server + '/admin/store/ResetTriplestoreContent?prependdefaults=false'

        req = requests.post(url,
                            headers={'Content-Type': 'application/json; charset=UTF-8'},
                            data=jsondata)
        self.on_api_error(req)
        res = req.json()
        #  pprint(res)
        return res


class Sipi:
    def __init__(self, sipiserver: str, token: str):
        self.sipiserver = sipiserver
        self.token = token

    def on_api_error(self, res):
        """
        Method to check for any API errors
        :param res: The input to check, usually JSON format
        :return: Possible KnoraError that is being raised
        """

        if (res.status_code != 200):
            raise KnoraError("SIPI-ERROR: status code=" + str(res.status_code) + "\nMessage:" + res.text)

        if 'error' in res:
            raise KnoraError("SIPI-ERROR: API error: " + res.error)

    def upload_image(self, filepath):
        files = {
            'file': (filepath, open(filepath, 'rb')),
        }
        req = requests.post(self.sipiserver + "/upload?token=" + self.token,
                            files=files)
        self.on_api_error(req)
        res = req.json()
        return res


class BulkImport:
    def __init__(self, schema: Dict):
        self.schema = schema
        self.proj_prefix = 'p' + schema['shortcode'] + '-' + schema["ontoname"]
        self.proj_iri = "http://api.knora.org/ontology/" + schema['shortcode'] + "/" + schema[
            "ontoname"] + "/xml-import/v1#"
        self.xml_prefixes = {
            None: self.proj_iri,
            "xsi": "http://www.w3.org/2001/XMLSchema-instance",
            self.proj_prefix: self.proj_iri,
            "knoraXmlImport": "http://api.knora.org/ontology/knoraXmlImport/v1#"
        }
        self.root = etree.Element('{http://api.knora.org/ontology/knoraXmlImport/v1#}resources',
                                  nsmap=self.xml_prefixes)
        self.project_shortcode = schema["shortcode"]

    def new_xml_element(self, tag: str, options: Dict = None, value: str = None):
        tagp = tag.split(':')
        if len(tagp) > 1:
            fulltag = '{' + self.xml_prefixes.get(tagp[0]) + '}' + tagp[1]
        else:
            fulltag = tagp[0]
        if options is None:
            ele = etree.Element(fulltag)
        else:
            ele = etree.Element(fulltag, options)
        if value is not None:
            ele.text = value
        return ele

    def write_xml(self, filename: str):
        # print(etree.tostring(self.root, pretty_print=True, xml_declaration=True, encoding='utf-8'))
        f = open(filename, "wb")
        f.write(etree.tostring(self.root, pretty_print=True, xml_declaration=True, encoding='utf-8'))
        f.close()

    def get_xml_string(self):
        """
        This method returns the Bulk-Import XML as an UTF-8 encoded string.
        :return: UTF-8 encoded string.
        """
        string = etree.tostring(self.root, pretty_print=True, xml_declaration=True, encoding='utf-8')
        return string

    def upload(self, user, password, hostname, port):
        """
        Upload the Bulk-Import XML to the server.
        :param user: the email of the user
        :param password: the password of the user
        :param hostname: the hostname (e.g., localhost, api.example.org, etc.)
        :param port: the port where the API is running (e.g., 3333)
        :return: the JSON response
        """
        project_iri = "http://rdfh.ch/projects/" + self.project_shortcode
        url_encoded_project_iri = urllib.parse.quote_plus(project_iri)
        bulkimport_api_url = "http://" + hostname + ":" + port + "/v1/resources/xmlimport/" + url_encoded_project_iri
        headers = {"Content-Type": "application/xml"}
        r = requests.post(bulkimport_api_url, data=self.get_xml_string(), headers=headers, auth=(user, password))
        return r.json()

    def add_resource(self, resclass: str, id: str, label: str, properties: Dict):
        """

        :param resclass:
        :param id:
        :param label:
        :param properties:
        :return:
        """

        def find_list_node_id(nodename: str, nodes: List):
            """
            Finds a list node ID from the nodename in a (eventually hierarchical) list of nodes

            :param nodename: Name of the node
            :param nodes: List of nodes
            :return: the id of the list node (an IRI)
            """
            for node in nodes:
                if node["name"] == nodename:
                    return node["id"]
                if node.get("nodes") is not None and len(node.get("nodes")) > 0:
                    node_id = find_list_node_id(nodename, node["nodes"])
                    if node_id is not None:
                        return node_id
            return None

        def process_properties(propinfo: Dict, valuestr: any):
            """
            Processes a property in order to generate the approptiate XML for V1 bulk import.

            :param pname: property name
            :param valuestr: value of the property
            :return: Tuple with xml options and processed value: (xmlopt, val)
            """
            switcher = {
                'TextValue': {'knoraType': 'richtext_value'},
                'ColorValue': {'knoraType': 'color_value'},
                'DateValue': {'knoraType': 'date_value'},
                'DecimalValue': {'knoraType': 'decimal_value'},
                'GeomValue': {'knoraType': 'geom_value'},
                'GeonameValue': {'knoraType': 'geoname_value'},
                'IntValue': {'knoraType': 'int_value'},
                'BooleanValue': {'knoraType': 'boolean_value'},
                'UriValue': {'knoraType': 'uri_value'},
                'IntervalValue': {'knoraType': 'interval_value'},
                'ListValue': {'knoraType': 'hlist_value'},
                'LinkValue': {'knoraType': 'link_value'}
            }
            for link_otype in self.schema["link_otypes"]:
                switcher[link_otype] = {'knoraType': 'link_value'}
            xmlopt = switcher.get(propinfo["otype"])
            if xmlopt is None:
                raise KnoraError("Did not find " + propinfo["otype"] + " in switcher!")
            if xmlopt['knoraType'] == 'link_value':
                xmlopt['target'] = str(valuestr)
                if validators.url(str(valuestr)):
                    xmlopt['linkType'] = 'iri'
                else:
                    xmlopt['linkType'] = 'ref'
                value = None
            elif propinfo["otype"] == 'ListValue':
                if validators.url(str(valuestr)):
                    # it's a full IRI identifying the node
                    value = valuestr
                else:
                    # it's only a node name. First let's get the node list from the ontology schema
                    list_id = propinfo["attr"]["hlist"]
                    for listname in self.schema["lists"]:
                        if self.schema["lists"][listname]["id"] == list_id:
                            nodes = self.schema["lists"][listname]["nodes"]
                    value = find_list_node_id(str(valuestr), nodes)
                # if value == 'http://rdfh.ch/lists/0808/X6bb-JerQyu5ULruCGEO0w':
                #     print("BANG!")
                #     exit(0)
            elif propinfo["otype"] == 'DateValue':
                # processing and validating date format
                res = re.match('(GREGORIAN:|JULIAN:)?(\d{4})?(-\d{1,2})?(-\d{1,2})?(:\d{4})?(-\d{1,2})?(-\d{1,2})?',
                               str(valuestr))
                if res is None:
                    raise KnoraError("Invalid date format! " + str(valuestr))
                dp = res.groups()
                calendar = 'GREGORIAN:' if dp[0] is None else dp[0]
                y1 = None if dp[1] is None else int(dp[1].strip('-: '))
                m1 = None if dp[2] is None else int(dp[2].strip('-: '))
                d1 = None if dp[3] is None else int(dp[3].strip('-: '))
                y2 = None if dp[4] is None else int(dp[4].strip('-: '))
                m2 = None if dp[5] is None else int(dp[5].strip('-: '))
                d2 = None if dp[6] is None else int(dp[6].strip('-: '))
                if y1 is None:
                    raise KnoraError("Invalid date format! " + str(valuestr))
                if y2 is not None:
                    date1 = y1 * 10000;
                    if m1 is not None:
                        date1 += m1 * 100
                    if d1 is not None:
                        date1 += d1
                    date2 = y2 * 10000;
                    if m2 is not None:
                        date2 += m2 * 100
                    if d1 is not None:
                        date2 += d2
                    if date1 > date2:
                        y1, y2 = y2, y1
                        m1, m2 = m2, m1
                        d1, d2 = d2, d1
                value = calendar + str(y1)
                if m1 is not None:
                    value += f'-{m1:02d}'
                if d1 is not None:
                    value += f'-{d1:02d}'
                if y2 is not None:
                    value += f':{y2:04d}'
                if m2 is not None:
                    value += f'-{m2:02d}'
                if d2 is not None:
                    value += f'-{d2:02d}'
            else:
                value = str(valuestr)
            return xmlopt, value

        if self.schema["resources"].get(resclass) is None:
            raise KnoraError('Resource class is not defined in ontology!')
        resnode = self.new_xml_element(self.proj_prefix + ':' + resclass, {'id': str(id)})

        labelnode = self.new_xml_element('knoraXmlImport:label')
        labelnode.text = str(label)
        resnode.append(labelnode)

        for prop_info in self.schema["resources"][resclass]:
            # first we check if the cardinality allows to add this property
            if properties.get(prop_info["propname"]) is None:  # this property-value is missing
                if prop_info["card"] == 'cardinality' \
                    and prop_info["cardval"] == 1:
                    raise KnoraError(
                        resclass + " requires exactly one " + prop_info["propname"] + "-value: none supplied!")
                if prop_info["card"] == 'minCardinality' \
                    and prop_info["cardval"] == 1:
                    raise KnoraError(
                        resclass + " requires at least one " + prop_info["propname"] + "-value: none supplied!")
                continue
            if type(properties[prop_info["propname"]]) is list:
                if len(properties[prop_info["propname"]]) > 1:
                    if prop_info["card"] == 'maxCardinality' \
                        and prop_info["cardval"] == 1:
                        raise KnoraError(
                            resclass + " allows maximal one " + prop_info["propname"] + "-value: several supplied!")
                    if prop_info["card"] == 'cardinality' \
                        and prop_info["cardval"] == 1:
                        raise KnoraError(
                            resclass + " requires exactly one " + prop_info["propname"] + "-value: several supplied!")
                for p in properties[prop_info["propname"]]:
                    xmlopt, value = process_properties(prop_info, p)
                    if xmlopt['knoraType'] == 'link_value':
                        pnode = self.new_xml_element(self.proj_prefix + ':' + prop_info["propname"])
                        pnode.append(self.new_xml_element(self.proj_prefix + ':' + prop_info["otype"], xmlopt, value))
                    else:
                        pnode = self.new_xml_element(self.proj_prefix + ':' + prop_info["propname"], xmlopt, value)
                    resnode.append(pnode)
            else:
                xmlopt, value = process_properties(prop_info, properties[prop_info["propname"]])
                if xmlopt['knoraType'] == 'link_value':
                    pnode = self.new_xml_element(self.proj_prefix + ':' + prop_info["propname"])
                    pnode.append(self.new_xml_element(self.proj_prefix + ':' + prop_info["otype"], xmlopt, value))
                else:
                    pnode = self.new_xml_element(self.proj_prefix + ':' + prop_info["propname"], xmlopt, value)
                resnode.append(pnode)
        self.root.append(resnode)


class IrisLookup:
    def __init__(self, local_id_to_iri_json):
        self.iris = local_id_to_iri_json

    def get_resource_iri(self, local_id):
        """
        Given the result of the bulk-import as json, allow retrieving the resource
        IRI based on the local ID.
        {'createdResources': [{'clientResourceID': 'LM_1',
                               'label': '1',
                               'resourceIri': 'http://rdfh.ch/0807/rNxoIK-oR_i0-lO21Y9-CQ'},
                              {'clientResourceID': 'LM_2']}

        :param local_id: the local id. resulting JSON from a bulk import upload.
        :return:
        """

        try:
            resources = self.iris["createdResources"]
            iri = ""
            for resource in resources:
                try:
                    res_id = resource["clientResourceID"]
                    if res_id == local_id:
                        iri = resource["resourceIri"]
                    else:
                        pass
                except KeyError:
                    pass

            if iri == "":
                return None
            else:
                return iri
        except KeyError:
            print("IrisLookup.get_resource_iri - 'createdResources' not found")

    def get_iris_json(self):
        return self.iris


class ListsLookup:
    def __init__(self, lists_json):
        self.lists = lists_json

    def get_list_iri(self, listname):
        return self.lists[listname]["id"]

    def get_list_node_iri(self, listname, nodename):
        if nodename is not None:
            nodes = self.lists[listname]["nodes"]
            res = ""
            for node in nodes:
                try:
                    res = node[nodename]["id"]
                except KeyError:
                    pass

            if res == "":
                return None
            else:
                return res
        else:
            return None

    def get_lists_json(self):
        return self.lists


if __name__ == '__main__':
    con = Knora('http://localhost:3333')
    con.login('root@example.com', 'test')
    res = con.get_resource_by_label('Bertschy, Leon',
                                    res_class="http://0.0.0.0:3333/ontology/0807/mls/v2#Lemma")
    print('RES-IRI: ', res['@id'])
    con.logout()




import unittest

from dsplib.models.connection import Connection
from dsplib.models.langstring import Languages, LangStringParam, LangString
from dsplib.models.listnode import ListNode


class TestListNode(unittest.TestCase):
    project = 'http://rdfh.ch/projects/0001'

    def test_ListNode_read(self):
        """
        Read an existing node
        :return: None
        """
        con = Connection('http://0.0.0.0:3333')
        con.login('root@example.com', 'test')
        node = ListNode(
            con=con,
            id='http://rdfh.ch/lists/0001/otherTreeList'
        ).read()
        self.assertEqual(node.id, 'http://rdfh.ch/lists/0001/otherTreeList')
        self.assertEqual(node.project, 'http://rdfh.ch/projects/0001')
        self.assertEqual(node.label['en'], 'Tree list root')
        self.assertTrue(node.isRootNode)

    def test_ListNode_read2(self):
        """
        read another existing node
        :return:
        """
        con = Connection('http://0.0.0.0:3333')
        con.login('root@example.com', 'test')
        node = ListNode(
            con=con,
            id='http://rdfh.ch/lists/0001/otherTreeList03'
        ).read()
        self.assertEqual(node.id, 'http://rdfh.ch/lists/0001/otherTreeList03')
        self.assertEqual(node.label['en'], 'Other Tree list node 03')
        self.assertFalse(node.isRootNode)
        self.assertEqual(node.rootNodeIri, 'http://rdfh.ch/lists/0001/otherTreeList')
        self.assertIsNone(node.children)

    def test_ListNode_create(self):
        con = Connection('http://0.0.0.0:3333')
        con.login('root@example.com', 'test')
        node = ListNode(
            con=con,
            project=self.project,
            label=LangString({Languages.DE: "root node 1"}),
            comments=LangString({Languages.DE: "first root node"}),
            name="test_node_1"
        ).create()
        self.assertIsNotNone(node.id)
        self.assertEqual(node.project, self.project)
        self.assertEqual(node.label['de'], "root node 1")
        self.assertEqual(node.comments['de'], "first root node")
        self.assertEqual(node.name, "test_node_1")
        self.assertTrue(node.isRootNode)

    def test_ListNode_hierarchy(self):
        """
        Create a node and a sub-node
        :return: None
        """
        con = Connection('http://0.0.0.0:3333')
        con.login('root@example.com', 'test')
        node = ListNode(
            con=con,
            project=self.project,
            label=LangString({Languages.DE: "root node 2"}),
            comments=LangString({Languages.DE: "second root node"}),
            name="test_node_2"
        ).create()
        node2 = ListNode(
            con=con,
            project=self.project,
            label=LangString({Languages.DE: 'Eine Knoten der Liste'}),
            comments=LangString({Languages.DE: "So ein Kommentar"}),
            name="NODE2",
            parent=node
        ).create()
        self.assertEqual(node2.label['de'], 'Eine Knoten der Liste')
        self.assertEqual(node2.comments['de'], "So ein Kommentar")
        self.assertEqual(node2.name, "NODE2")
        self.assertFalse(node2.isRootNode)

    def test_ListNode_update(self):
        """
        Update the data of a node...
        :return:
        """
        con = Connection('http://0.0.0.0:3333')
        con.login('root@example.com', 'test')
        node = ListNode(
            con=con,
            project=self.project,
            label=LangString({Languages.DE: "root node 3"}),
            comments=LangString({Languages.DE: "Third root node"}),
            name="test_node_3"
        ).create()
        node.addLabel('fr', 'Une racine d\' une liste')
        node.rmLabel('de')
        node.addComment('fr', 'un commentaire en français')
        node.rmComment('de')
        node.name = 'GAGAGA'
        node.update()
        self.assertEqual(node.label['fr'], 'Une racine d\' une liste')
        self.assertEqual(node.comments['fr'], 'un commentaire en français')
        self.assertEqual(node.name, 'GAGAGA')

    def test_ListNode_getAllLists(self):
        """
        gett all root nodes, that is all lists
        :return: None
        """
        con = Connection('http://0.0.0.0:3333')
        con.login('root@example.com', 'test')
        lists = ListNode.getAllLists(con, 'http://rdfh.ch/projects/0001')
        tmp = list(map(lambda a: a.id, lists))
        self.assertIn('http://rdfh.ch/lists/0001/otherTreeList', tmp)
        self.assertIn('http://rdfh.ch/lists/0001/treeList', tmp)

    def test_ListNode_getAllNodes(self):
        """
        Get all node of a list
        :return: None
        """
        con = Connection('http://0.0.0.0:3333')
        con.login('root@example.com', 'test')
        root = ListNode(
            con=con,
            id='http://rdfh.ch/lists/0001/otherTreeList'
        ).getAllNodes()
        self.assertTrue(root.isRootNode)
        self.assertEqual(root.project, 'http://rdfh.ch/projects/0001')
        self.assertEqual(root.label['en'], 'Tree list root')
        self.assertIsNotNone(root.children)

        self.assertEqual(root.children[0].id, 'http://rdfh.ch/lists/0001/otherTreeList01')
        self.assertEqual(root.children[0].name, 'Other Tree list node 01')
        self.assertEqual(root.children[0].label['en'], 'Other Tree list node 01')

        self.assertEqual(root.children[1].id, 'http://rdfh.ch/lists/0001/otherTreeList02')
        self.assertEqual(root.children[1].name, 'Other Tree list node 02')
        self.assertEqual(root.children[1].label['en'], 'Other Tree list node 02')

        self.assertEqual(root.children[2].id, 'http://rdfh.ch/lists/0001/otherTreeList03')
        self.assertEqual(root.children[2].name, 'Other Tree list node 03')
        self.assertEqual(root.children[2].label['en'], 'Other Tree list node 03')

        self.assertIsNotNone(root.children[2].children)
        self.assertEqual(root.children[2].children[0].id, 'http://rdfh.ch/lists/0001/otherTreeList10')
        self.assertEqual(root.children[2].children[0].name, 'Other Tree list node 10')
        self.assertEqual(root.children[2].children[0].label['en'], 'Other Tree list node 10')

        self.assertEqual(root.children[2].children[1].id, 'http://rdfh.ch/lists/0001/otherTreeList11')
        self.assertEqual(root.children[2].children[1].name, 'Other Tree list node 11')
        self.assertEqual(root.children[2].children[1].label['en'], 'Other Tree list node 11')

        #self.assertTrue(False)


if __name__ == '__main__':
    unittest.main()



import os
import re

from typing import List, Set, Dict, Tuple, Optional, Union
from lxml import etree
from pprint import pprint

from dsplib.models.connection import Connection
from dsplib.models.group import Group
from dsplib.models.project import Project
from dsplib.models.resource import ResourceInstanceFactory
from dsplib.models.value import BooleanValue, ColorValue, DateValue, DecimalValue, IntValue, IntervalValue, TextValue, \
    UriValue, KnoraStandoffXml, make_value
from dsplib.models.permission import PermissionValue, Permissions
from dsplib.models.sipi import Sipi

StrDict = Dict[str, str]

StrObj = Union[str, StrDict]

VarStrObj = Union[StrObj, List[StrObj]]

richtext_tags = [
    'p', 'em', 'strong', 'u', 'sub', 'strike', 'a', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'ol', 'ul', 'li', 'tbody',
    'table', 'tr', 'td', 'br', 'hr', 'pre', 'cite', 'blockquote', 'code'
]

class XmlError(BaseException):

    def __init__(self, msg: str):
        self._message = msg

    def __str__(self):
        return 'XML-ERROR: ' + self._message


class ProjectContext:
    _projects: Project
    _groups: Group
    _projectmap: Dict[str, str]
    _invprojectmap: Dict[str, str]
    _groupmap: Dict[str, str]
    _shortcode: Union[str, None]
    _project_name: Union[str, None]

    def __init__(self, con: Connection, shortcode: Optional[str] = None):
        self._shortcode = shortcode
        self._projects = Project.getAllProjects(con=con)
        self._projectmap: Dict[str, str] = {x.shortname: x.id for x in self._projects}
        invprojectmap: Dict[str, str] = {x.id: x.shortname for x in self._projects}
        self._groups = Group.getAllGroups(con=con)
        self._groupmap: Dict[str, str] = {invprojectmap[x.project] + ':' + x.name: x.id for x in self._groups}
        self._project_name = None
        if self._shortcode:
            for p in self._projects:
                if p.shortcode == self._shortcode:
                    self._project_name = p.shortname
                    break

    @property
    def groupmap(self):
        return self._groupmap

    @property
    def shortcode(self) -> Union[str, None]:
        return self._shortcode

    @shortcode.setter
    def shortcode(self, val: str) -> None:
        for p in self._projects:
            if p.shortcode == self._shortcode:
                self._project_name = p.shortname
                break

    @property
    def project_name(self) -> Union[str, None]:
        return self._project_name


class KnoraValue:
    _value: Union[str, KnoraStandoffXml]
    _resrefs: List[str]
    _comment: str
    _permissions: str
    is_richtext: bool

    def __init__(self,
                 node: etree.Element,
                 valtype: str,
                 listname: Optional[str] = None) -> None:

        self._resrefs = None
        self._comment = node.get('comment')
        self._permissions = node.get('permissions')
        if node.get('encoding') == 'xml':
            node.attrib.clear()
            xmlstr = etree.tostring(node, encoding="unicode", method="xml")
            xmlstr = xmlstr.replace('<text xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">', '')
            xmlstr = xmlstr.replace('</text>', '')
            self._value = KnoraStandoffXml(xmlstr)
            tmpidlist = self._value.findall()
            if tmpidlist:
                refs = set()
                for tmpid in tmpidlist:
                    refs.add(tmpid.split(':')[1])
                self._resrefs = list(refs)
        else:
            if valtype == 'list':
                self._value = listname + ':' + "".join(node.itertext())
            else:
                self._value = "".join(node.itertext())

    @property
    def value(self):
        return self._value

    @property
    def resrefs(self):
        return self._resrefs

    @property
    def comment(self):
        return self._comment

    @property
    def permissions(self):
        return self._permissions

    def print(self) -> None:
        """
        Print value to stdout for debugging...

        :return: None
        """
        print('    Value: ' + str(self._value))
        if self._comment:
            print('   Comment:' + self._comment)
        if self._resrefs is not None:
            for i in self._resrefs:
                print('    resref: ' + i)


class KnoraProperty:
    _name: str
    _valtype: str
    _values: List[KnoraValue]

    def __init__(self, node: etree.Element, valtype: str, default_ontology: Optional[str] = None):
        tmp = node.attrib['name'].split(':')
        if len(tmp) > 1:
            if tmp[0]:
                self._name = node.attrib['name']
            else:
                self._name = default_ontology + ':' + tmp[1]
        else:
            self._name = 'knora-admin:' + tmp[0]
        listname = node.attrib.get('list')  # safe the list name if given (only for lists)
        self._valtype = valtype
        self._values = []

        for subnode in node:
            if subnode.tag == valtype:  # the subnode must correspond to the expected value type
                self._values.append(KnoraValue(subnode, valtype, listname))
            else:
                raise XmlError('Unexpected tag: "{}" <property> may contain only <value> tags!'.format(subnode.tag))

    @property
    def name(self):
        return self._name

    @property
    def valtype(self):
        return self._valtype

    @property
    def values(self):
        return self._values

    def print(self) -> None:
        print('  Property: {} Type: {}'.format(self._name, self._valtype))
        for value in self._values:
            value.print()


class KnoraResource:
    _id: str
    _label: str
    _restype: str
    _permissions: str
    _bitstream: str
    _properties: List[KnoraProperty]

    def __init__(self, node: etree.Element, default_ontology: Optional[str] = None) -> None:
        """
        Constructor that parses a resource node from the XML DOM

        :param context: Context for DOM node traversal
        :param node: The DOM node to be processed (representing a resource)
        """
        self._id = node.attrib['id']  # safe the unique id
        self._label = node.attrib['label']
        tmp = node.attrib['restype'].split(':')
        if len(tmp) > 1:
            if tmp[0]:
                self._restype = node.attrib['restype']
            else:
                self._restype = default_ontology + ':' + tmp[1]
        else:
            self._restype = 'knora-admin:' + tmp[0]
        self._permissions = node.attrib['permissions']
        self._bitstream = None
        self._properties = []
        for subnode in node:
            if subnode.tag == 'bitstream':
                self._bitstream = subnode.text
            elif subnode.tag is etree.Comment:
                    continue
            else:
                ptype, dummy = subnode.tag.split('-')
                self._properties.append(KnoraProperty(subnode, ptype, default_ontology))

    @property
    def id(self) -> str:
        return self._id

    @property
    def label(self) -> str:
        return self._label

    @property
    def restype(self) -> str:
        return self._restype

    @property
    def bitstream(self) -> str:
        return self._bitstream

    @property
    def permissions(self):
        return self._permissions

    def print(self):
        print(f'Resource: id={self._id} restype: {self._restype} label: {self._label}')
        if self._bitstream is not None:
            print(' Bitstream: ' + self._bitstream)
        for property in self._properties:
            property.print()

    def get_resptrs(self) -> List[str]:
        """
        Return a list of all reesource id's  that a referenced by this resource
        :return: List of resources identified by their unique id's
        """
        resptrs: List[str] = []
        for property in self._properties:
            if property.valtype == 'resptr':
                for value in property.values:
                    resptrs.append(value.value)
            elif property.valtype == 'text':
                for value in property.values:
                    if value.resrefs is not None:
                        resptrs.extend(value.resrefs)
        return resptrs

    def get_propvals(self,
                     resiri_lookup: StrDict,
                     permissions_lookup: StrDict) -> Dict[str, VarStrObj]:
        """
        A function which retrieves...

        :param resiri_lookup: Is used to solve internal unique_id's of resourcs to real IRI's
        :param permissions_lookup: Is usd to resolve thee permission ID's to permission sets
        :return: A dict of values with the property name as key and a single value. This dict represents
                 the JSON structure that Knora.create_resource() expects.
        """
        propdata = {}
        for property in self._properties:
            vals: List[StrObj] = []  # == List[Union[str,StrDict]
            for value in property.values:
                v: str
                if property.valtype == 'resptr':  # we have a resptr, therefore simple lookup or IRI
                    iri = resiri_lookup.get(value.value)
                    if iri is not None:
                        v = iri
                    else:
                        v = value.value  # if we do not find the unique_id, we assume it's a valid knora IRI
                elif property.valtype == 'text':
                    if isinstance(value.value, KnoraStandoffXml):
                        irirefs = value.value.findall()  # The IRI's must be embedded  as "...IRI:unique_id:IRI..."
                        for iriref in irirefs:
                            resid = iriref.split(':')[1]
                            iri = resiri_lookup.get(resid)
                            value.value.replace(iriref, iri)
                    v = value.value
                else:
                    v = value.value

                if value.comment is None and value.permissions is None:
                    # no comment or permissions
                    vals.append(v)
                else:
                    # we have comment or permissions
                    tmp = {'value': v}
                    if value.comment is not None:
                        tmp['comment'] = value.comment
                    if value.permissions is not None:
                        tmp['permissions'] = permissions_lookup.get(value.permissions)
                    vals.append(tmp)
            propdata[property.name] = vals if len(vals) > 1 else vals[0]  # append a Union[StrObj,List[StrObj]]
        return propdata


class XmlAllow:
    _group: str
    _permission: str

    def __init__(self, node: etree.Element, project_context: ProjectContext) -> None:
        """
        Constructor which parses the XML DOM allow element

        :param context: Context for DOM node traversal
        :param node: The DOM node to be processed (representing an single right in a permission set)
        """
        tmp = node.attrib['group'].split(':')
        sysgroups = ['UnknownUser', 'KnownUser', 'ProjectMember', 'Creator', 'ProjectAdmin', 'SystemAdmin']
        if len(tmp) > 1:
            if tmp[0]:
                if tmp[0] == 'knora-admin' and tmp[1] in sysgroups:
                    self._group = node.attrib['group']
                else:
                    self._group = project_context.groupmap.get(node.attrib['group'])
                    if self._group is None:
                        raise XmlError("Group \"{}\" is not known: Cannot find project!".format(node.attrib['group']))
            else:
                if project_context.project_name is None:
                    raise XmlError("Project shortcode has not been set in ProjectContext")
                self._group = project_context.project_name + ':' + tmp[1]
        else:
            if tmp[0] in sysgroups:
                self._group = 'knora-admin:' + node.attrib['group']
            else:
                raise XmlError("Group \"{}\" is not known: ".format(node.attrib['group']))
        self._permission = node.text

    @property
    def group(self):
        return self._group

    @property
    def permission(self):
        return self._permission

    def print(self):
        print("  group=", self._group, " permission=", self._permission)


class XmlPermission:
    """
    A class representing a permission set
    """
    _id: str
    _allows: List[XmlAllow]

    def __init__(self, node: etree.Element, project_context: ProjectContext) -> None:
        """
        Constructor which parses a XML DOM permissions element representing an named permission set

        :param context: Context for DOM node traversal
        :param node: The DOM node to be processed (representing an a permission set)
        """
        self._allows = []
        self._id = node.attrib['id']
        for allow_node in node:
            self._allows.append(XmlAllow(allow_node, project_context))

    @property
    def id(self) -> str:
        return self._id

    @property
    def allows(self) -> List[XmlAllow]:
        return self._allows

    def get_permission_instance(self) -> Permissions:
        permissions = Permissions()
        for allow in self._allows:
            permissions.add(allow.permission, allow.group)
        return permissions

    def __str__(self):
        allowstrs: List[str] = []
        for allow in self._allows:
            allowstrs.append("{} {}".format(allow.permission, allow.group))
        return '|'.join(allowstrs)

    def print(self):
        print('Permission: ', self._id)
        for a in self._allows:
            a.print()


def do_sortorder(resources: List[KnoraResource]) -> List[KnoraResource]:
    """
    Sort the list of resources such that resources that reference other resources are
    added after the referenced resources. It will fail with an error if there are circular
    references.

    :param resources: List of resources before sorting
    :return: Sorted list of resources
    """
    #
    # here we sort the resources according to outgoing resptrs
    #
    ok_resources: [KnoraResource] = []
    notok_resources: [KnoraResource] = []
    ok_resids : [str] = []
    cnt = 0
    notok_len = 9999999
    while len(resources) > 0 and cnt < 10000:
        for resource in resources:
            resptrs = resource.get_resptrs()
            if len(resptrs) == 0:
                ok_resources.append(resource)
                ok_resids.append(resource.id)
            else:
                ok = True
                for resptr in resptrs:
                    if resptr in ok_resids:
                        pass
                    else:
                        ok = False;
                if ok:
                    ok_resources.append(resource)
                    ok_resids.append(resource.id)
                else:
                    notok_resources.append(resource)
        resources = notok_resources
        if not len(notok_resources) < notok_len:
            print('Cannot resolve resptr dependencies. Giving up....')
            print(len(notok_resources))
            for r in notok_resources:
                print('Resource {} has unresolvable resptrs to: '.format(r.id), end=' ')
                for x in r.get_resptrs():
                    print(x, end=' ')
                print('')
                print('=============')
            exit(5)
        notok_len = len(notok_resources)
        notok_resources = []
        cnt += 1
        print('{}. Ordering pass Finished!'.format(cnt))
    print('Remaining: {}'.format(len(resources)))
    return ok_resources


def xml_upload(input_file: str,
               server: str,
               user: str,
               password: str,
               imgdir: str,
               sipi: str,
               verbose: bool,
               validate: bool) -> bool:
    current_dir = os.path.dirname(os.path.realpath(__file__))

    xmlschema_doc = etree.parse(os.path.join(current_dir, 'knora-data-schema.xsd'))
    xmlschema = etree.XMLSchema(xmlschema_doc)
    doc = etree.parse(input_file)
    xmlschema.assertValid(doc)

    del xmlschema
    del doc
    del xmlschema_doc

    if validate:
        return

    print("The input data file is syntactically correct and passed validation!")

    #
    # Connect to the DaSCH Service Platform API
    #
    con = Connection(server)
    con.login(user, password)

    proj_context = ProjectContext(con=con)

    resources: List[KnoraResource] = []
    permissions: Dict[str, XmlPermission] = {}
    shortcode: Union[str, None] = None
    default_ontology = None

    #
    # read the XML file containing the data, including project shortcode
    #
    tree = etree.parse(input_file)
    knora = tree.getroot()
    default_ontology = knora.attrib['default-ontology']
    shortcode = knora.attrib['shortcode']
    for child in knora:
        if child.tag == "permissions":
            permission = XmlPermission(child, proj_context)
            permissions[permission.id] = permission
        elif child.tag == "resource":
            resources.append(KnoraResource(child, default_ontology))

    #
    # sort the resources so that resources which do not link to others come first
    #
    resources = do_sortorder(resources)

    sipi = Sipi(sipi, con.get_token())

    factory = ResourceInstanceFactory(con, shortcode)

    permissions_lookup: Dict[str, Permissions] = {}
    for key, perm in permissions.items():
        permissions_lookup[key] = perm.get_permission_instance()

    resclassnames = factory.get_resclass_names()
    resclasses: Dict[str, type] = {}
    for resclassname in resclassnames:
        resclasses[resclassname] = factory.get_resclass(resclassname)
    resiri_lookup: StrDict = {}

    for resource in resources:
        #resource.print()
        if resource.bitstream:
            img = sipi.upload_bitstream(os.path.join(imgdir, resource.bitstream))
            bitstream = img['uploadedFiles'][0]['internalFilename']
        else:
            bitstream = None
        instance = resclasses[resource.restype](con=con,
                                                label=resource.label,
                                                permissions=permissions_lookup.get(resource.permissions),
                                                bitstream=bitstream,
                                                values=resource.get_propvals(resiri_lookup, permissions_lookup)).create()
        resiri_lookup[resource.id] = instance.iri
        print("Created:", instance.iri)



from typing import List, Set, Dict, Tuple, Optional
from openpyxl import load_workbook, worksheet

from ..models.connection import Connection
from ..models.project import Project
from ..models.listnode import ListNode
from ..models.langstring import Languages, LangString
from ..models.helpers import BaseError

from pprint import pprint

def list_creator(con: Connection, project: Project, parent_node: ListNode, nodes: List[dict]):
    nodelist = []
    for node in nodes:
        newnode = ListNode(
            con=con,
            project=project,
            label=node["labels"],
            comments=node.get("comments"),
            name=node["name"],
            parent=parent_node
        ).create()
        if node.get('nodes') is not None:
            subnodelist = list_creator(con, project, newnode, node['nodes'])
            nodelist.append({newnode.name: {"id": newnode.id, 'nodes': subnodelist}})
        else:
            nodelist.append({newnode.name: {"id": newnode.id}})
    return nodelist


def validate_list_from_excel(filepath: str,
                             sheetname: str,
                             startrow: int = 1,
                             startcol: int = 1,
                             verbose: bool = False):

    def analyse_level(ws: worksheet, level: int, row: int, col: int, preval: List[str]) ->int:
        cell = ws.cell(column=col, row=row)
        if col > 1:
            preval.append(ws.cell(column=col - 1, row=row).value)
        while cell.value:
            for idx, val in enumerate(preval[:-1]):
                if val != ws.cell(column=idx + 1, row=row).value:
                    raise BaseError(f"Inconsistency in Excel list! {val} not equal {ws.cell(column=idx + 1, row=row).value}")
            if ws.cell(column=col + 1, row=row).value:
                row = analyse_level(ws=ws, level=level + 1, col=col + 1, row=row, preval=preval)
                if not ws.cell(column=col, row=row).value:
                    if col > 1:
                        preval.pop()
                    return row
            else:
                if verbose:
                    print(f"Node on level={level}, value={cell.value} ok...")
            row += 1
            cell = ws.cell(column=col, row=row)
        if col > 1:
            preval.pop()
        return row - 1

    wb = load_workbook(filename=filepath, read_only=True)
    ws = wb[sheetname]
    tmp = []
    analyse_level(ws=ws, level=1, row=startrow, col=startcol, preval=tmp)


def json_list_from_excel(rootnode: {}, filepath: str, sheetname: str, startrow: int = 1, startcol: int = 1):

    names: Set[str] = set()

    def analyse_level(ws: worksheet, parentnode: {}, row: int, col: int, preval: List[str]) ->int:
        nodes: [] = []
        currentnode: {}

        cell = ws.cell(column=col, row=row)
        if col > 1:
            preval.append(ws.cell(column=col - 1, row=row).value)
        while cell.value:
            for idx, val in enumerate(preval[:-1]):
                if val != ws.cell(column=idx + 1, row=row).value:
                    raise BaseError(f"Inconsistency in Excel list! {val} not equal {ws.cell(column=idx + 1, row=row).value}")
            if ws.cell(column=col + 1, row=row).value:
                row = analyse_level(ws=ws, parentnode=currentnode, col=col + 1, row=row, preval=preval)
                if not ws.cell(column=col, row=row).value:
                    if col > 1:
                        preval.pop()
                    parentnode["nodes"] = nodes
                    return row
            else:
                tmpstr = cell.value
                tmpstr = tmpstr.split(" ")
                tmpstr = [w.title() for w in tmpstr]
                tmpstr = "".join(tmpstr)
                tmpstr = tmpstr[0].lower() + tmpstr[1:]
                while tmpstr in names:
                    tmpstr = tmpstr + "_"
                names.add(tmpstr)
                currentnode = {
                    "name": tmpstr,
                    "labels": {"en": cell.value}
                }
                nodes.append(currentnode)
                print(f"Adding list node: value={cell.value}")
            row += 1
            cell = ws.cell(column=col, row=row)
        if col > 1:
            preval.pop()
        parentnode["nodes"] = nodes
        return row - 1

    wb = load_workbook(filename=filepath, read_only=True)
    ws = wb[sheetname]
    tmp = []
    analyse_level(ws=ws, parentnode=rootnode, row=startrow, col=startcol, preval=tmp)





import sys
import unittest
import pprint
import os

sys.path.append("../knora")

from dsplib.models.helpers import BaseError
from dsplib.models.connection import Connection


class TestConnection(unittest.TestCase):
    def test_Connection(self):
        con = Connection('http://0.0.0.0:3333')
        self.assertIsInstance(con, Connection)

    def test_loginout(self):
        con = Connection('http://0.0.0.0:3333')
        con.login('root@example.com', 'test')
        self.assertIsNotNone(con.token)
        con.logout()
        self.assertIsNone(con.token)
        self.assertRaisesRegex(BaseError, 'KNORA-ERROR: status code=400*', con.login, 'invalid', 'invalid')

    def test_get(self):
        con = Connection('http://0.0.0.0:3333')
        res = con.get("/ontology/0001/anything/simple/v2")
        con.logout()
        self.assertIsNotNone(res['@graph'])
        self.assertRaises(BaseError, con.get, "/gagaga")

    def test_put(self):
        pass

    def test_post(self):
        resinfo = """{
            "@type" : "anything:Thing",
            "knora-api:attachedToProject" : {
                "@id" : "http://rdfh.ch/projects/0001"
            },
            "anything:hasBoolean" : {
                "@type" : "knora-api:BooleanValue",
                "knora-api:booleanValueAsBoolean" : true
            },
            "rdfs:label" : "knora-py thing",
            "knora-api:hasPermissions" : "CR knora-admin:Creator|V http://rdfh.ch/groups/0001/thing-searcher",
            "@context" : {
                "rdf" : "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
                "knora-api" : "http://api.knora.org/ontology/knora-api/v2#",
                "rdfs" : "http://www.w3.org/2000/01/rdf-schema#",
                "xsd" : "http://www.w3.org/2001/XMLSchema#",
                "anything" : "http://0.0.0.0:3333/ontology/0001/anything/v2#"
            }
        }
        """
        con = Connection('http://0.0.0.0:3333')
        con.login('root@example.com', 'test')
        res = con.post('/v2/resources', resinfo)
        self.assertIsNotNone(res['@id'])
        id = res['@id']
        self.assertEqual(res['@type'], 'anything:Thing')
        self.assertEqual(res['rdfs:label'], 'knora-py thing')

        eraseinfo = """{
            "@id" : "%s",
            "@type" : "anything:Thing",
            "@context" : {
                "rdf" : "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
                "knora-api" : "http://api.knora.org/ontology/knora-api/v2#",
                "rdfs" : "http://www.w3.org/2000/01/rdf-schema#",
                "xsd" : "http://www.w3.org/2001/XMLSchema#",
                "anything" : "http://0.0.0.0:3333/ontology/0001/anything/v2#"
            }
        }
        """ % id
        res = con.post('/v2/resources/erase', eraseinfo)
        self.assertIsNotNone(res['knora-api:result'])
        self.assertEqual(res['knora-api:result'], 'Resource erased')

    def test_delete(self):
        pass


if __name__ == '__main__':
    unittest.main()



import os
from typing import List, Set, Dict, Tuple, Optional
import json
from jsonschema import validate

from ..models.helpers import Actions, BaseError, Context, Cardinality
from .onto_commons import list_creator, validate_list_from_excel, json_list_from_excel


def list_excel2json(excelpath: str,
                    sheetname: str,
                    shortcode: str,
                    listname: str,
                    label: str,
                    lang: str,
                    outfile: str,
                    verbose: bool):
    current_dir = os.path.dirname(os.path.realpath(__file__))
    langs = ["en", "de", "fr", "it"]

    if lang not in langs:
        raise BaseError(f"Language '{lang}' not supported!")

    rootnode = {
        "name": listname,
        "labels": {
            lang: label
        }
    }

    json_list_from_excel(rootnode, excelpath, sheetname)
    jsonobj = {
        "project": {
            "shortcode": shortcode,
            "lists": [
                rootnode
            ]
        }
    }
    with open(os.path.join(current_dir, 'knora-schema-lists.json')) as s:
        schema = json.load(s)
    validate(jsonobj, schema)
    json.dump(jsonobj, open(outfile, "w"), indent=4)



import json
from pystrict import strict
from typing import List, Set, Dict, Tuple, Optional, Any, Union
from urllib.parse import quote_plus

from .helpers import Actions, BaseError
from .langstring import Languages, LangStringParam, LangString
from .connection import Connection
from .model import Model

from pprint import pprint

class SetEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)

"""
This module implements the handling (CRUD) of Knora projects.

CREATE:
    * Instantiate a new object of the class Project with all required parameters
    * Call the ``create``-method on the instance

READ:
    * Instantiate a new object with ``id``(IRI of project) given
    * Call the ``read``-method on the instance
    * Access the information that has been provided to the instance

UPDATE:
    * You need an instance of an existing Project by reading an instance
    * Change the attributes by assigning the new values
    * Call the ``update```method on the instance

DELETE
    * Instantiate a new objects with ``id``(IRI of project) given, or use any instance that has the id set
    * Call the ``delete``-method on the instance

In addition there is a static methods ``getAllProjects`` which returns a list of all projects
"""

@strict
class Project(Model):
    """
    This class represents a project in Knora.

    Attributes
    ----------

    con : Connection
        A Connection instance to a Knora server

    id : str
        IRI of the project [readonly, cannot be modified after creation of instance]

    shortcode : str
        Knora project shortcode [readonly, cannot be modified after creation of instance]

    shortname : str
        Knora project shortname [read/write]

    longname : str
        Knora project longname [read/write]

    description : LangString
        Knora project description in a given language (Languages.EN, Languages.DE, Languages.FR, Languages.IT).
        A desciption can be add/replaced or removed with the methods ``addDescription``and ``rmDescription``.

    keywords : Set[str]
        Set of keywords describing the project. Keywords can be added/removed by the methods ``addKeyword``
        and ``rmKeyword``

    ontologies : Set[str]
        Set if IRI's of the ontologies attached to the project [readonly]

    selfjoin : bool
        Boolean if the project allows selfjoin


    Methods
    -------

    create : Knora project information object
        Creates a new project and returns the information from the project as it is in Knora

    read : Knora project information object
        Read project data from an existing project

    update : Knora project information object
        Updates the changed attributes and returns the updated information from the project as it is in Knora

    delete : Knora result code
        Deletes a project and returns the result code

    getAllprojects [static]: List of all projects
        Returns a list of all projects available

    print : None
        Prints the project information to stdout

    """
    _id: str
    _shortcode: str
    _shortname: str
    _longname: str
    _description: LangString
    _keywords: Set[str]
    _ontologies: Set[str]
    _selfjoin: bool
    _status: bool
    _logo: Optional[str]

    SYSTEM_PROJECT: str = "http://www.knora.org/ontology/knora-admin#SystemProject"

    def __init__(self,
                 con:  Connection,
                 id: Optional[str] = None,
                 shortcode: Optional[str] = None,
                 shortname: Optional[str] = None,
                 longname: Optional[str] = None,
                 description: LangStringParam = None,
                 keywords: Optional[Set[str]] = None,
                 ontologies: Optional[Set[str]] = None,
                 selfjoin: Optional[bool] = None,
                 status: Optional[bool] = None,
                 logo: Optional[str] = None):
        """
        Constructor for Project

        :param con: Connection instance
        :param id: IRI of the project [required for CREATE, READ]
        :param shortcode: Shortcode of the project. String inf the form 'XXXX' where each X is a hexadezimal sign 0-1,A,B,C,D,E,F. [required for CREATE]
        :param shortname: Shortname of the project [required for CREATE]
        :param longname: Longname of the project [required for CREATE]
        :param description: LangString instance containing the description [required for CREATE]
        :param keywords: Set of keywords [required for CREATE]
        :param ontologies: Set of ontologies that belong to this project [optional]
        :param selfjoin: Allow selfjoin [required for CREATE]
        :param status: Status of project (active if True) [required for CREATE]
        :param logo: Path to logo image file [optional] NOT YET USED
        """
        super().__init__(con)
        self._id = id
        self._shortcode = shortcode
        self._shortname = shortname
        self._longname = longname
        self._description = LangString(description)
        self._keywords = keywords
        if not isinstance(ontologies, set) and ontologies is not None:
            raise BaseError('Ontologies must be a set of strings or None!')
        self._ontologies = ontologies
        self._selfjoin = selfjoin
        self._status = status
        self._logo = logo

    def __str__(self):
        tmpstr = self._id + "\n  " + self._shortcode + "\n  " + self._shortname
        return tmpstr

    #
    # Here follows a list of getters/setters
    #
    @property
    def id(self) -> Optional[str]:
        return self._id

    @id.setter
    def id(self, value: str) -> None:
        raise BaseError('Project id cannot be modified!')

    @property
    def shortcode(self) -> Optional[str]:
        return self._shortcode

    @shortcode.setter
    def shortcode(self, value: str) -> None:
        raise BaseError('Shortcode id cannot be modified!')

    @property
    def shortname(self) -> Optional[str]:
        return self._shortname

    @shortname.setter
    def shortname(self, value: str) -> None:
        if self._shortname != str(value):
            self._shortname = str(value)
            self._changed.add('shortname')

    @property
    def longname(self) -> Optional[str]:
        return self._longname

    @longname.setter
    def longname(self, value: str) -> None:
        if self._longname != str(value):
            self._longname = str(value)
            self._changed.add('longname')

    @property
    def description(self) -> Optional[LangString]:
        return self._description

    @description.setter
    def description(self, value: Optional[LangString]) -> None:
        self._description = LangString(value)
        self._changed.add('description')

    def addDescription(self, lang: Union[Languages, str], value: str) -> None:
        """
        Add/replace a project description with the given language (executed at next update)

        :param lang: The language the description is in, either a string "EN", "DE", "FR", "IT" or a Language instance
        :param value: The text of the description
        :return: None
        """

        self._description[lang] = value
        self._changed.add('description')

    def rmDescription(self, lang: Union[Languages, str]) -> None:
        """
        Remove a description from a project (executed at next update)

        :param lang: The language the description to be removed is in, either a string "EN", "DE", "FR", "IT" or a Language instance
        :return: None
        """

        del self._description[lang]
        self._changed.add('description')

    @property
    def keywords(self) -> Set[str]:
        return self._keywords

    @keywords.setter
    def keywords(self, value: Union[List[str], Set[str]]):
        if isinstance(value, set):
            self._keywords = value
            self._changed.add('keywords')
        elif isinstance(value, list):
            self._keywords = set(value)
            self._changed.add('keywords')
        else:
            raise BaseError('Must be a set of strings!')

    def addKeyword(self, value: str):
        """
        Add a new keyword to the set of keywords. (executed at next update)
        May raise a BaseError

        :param value: keyword
        :return: None
        """

        self._keywords.add(value)
        self._changed.add('keywords')

    def rmKeyword(self, value: str):
        """
        Remove a keyword from the list of keywords (executed at next update)
        May raise a BaseError

        :param value: keyword
        :return: None
        """
        try:
            self._keywords.remove(value)
        except KeyError as ke:
            raise BaseError('Keyword "'  + value + '" is not in keyword set')
        self._changed.add('keywords')

    @property
    def ontologies(self) -> Set[str]:
        return self._ontologies

    @ontologies.setter
    def ontologies(self, value: Set[str]) -> None:
        raise BaseError('Cannot add a ontology!')

    @property
    def selfjoin(self) -> Optional[bool]:
        return self._selfjoin

    @selfjoin.setter
    def selfjoin(self, value: bool) -> None:
        if self._selfjoin != value:
            self._changed.add('selfjoin')
            self._selfjoin = value

    @property
    def status(self) -> bool:
        return self._status

    @status.setter
    def status(self, value: bool) -> None:
        if self._status != value:
            self._status = value
            self._changed.add('status')

    @property
    def logo(self) -> str:
        return self._logo

    @logo.setter
    def logo(self, value: str) -> None:
        if self._logo != value:
            self._logo = value
            self._changed.add('logo')

    @classmethod
    def fromJsonObj(cls, con: Connection, json_obj: Any) -> Any:
        """
        Internal method! Should not be used directly!

        This method is used to create a Project instance from the JSON data returned by Knora

        :param con: Connection instance
        :param json_obj: JSON data returned by Knora as python3 object
        :return: Project instance
        """
        id = json_obj.get('id')
        if id is None:
            raise BaseError('Project id is missing')
        shortcode = json_obj.get('shortcode')
        if shortcode is None:
            raise BaseError("Shortcode is missing")
        shortname = json_obj.get('shortname')
        if shortname is None:
            raise BaseError("Shortname is missing")
        longname = json_obj.get('longname')
        if longname is None:
            raise BaseError("Longname is missing")
        description = LangString.fromJsonObj(json_obj.get('description'))
        keywords = set(json_obj.get('keywords'))
        if keywords is None:
            raise BaseError("Keywords are missing")
        ontologies = set(json_obj.get('ontologies'))
        if ontologies is None:
            raise BaseError("ontologies are missing")
        selfjoin = json_obj.get('selfjoin')
        if selfjoin is None:
            raise BaseError("Selfjoin is missing")
        status = json_obj.get('status')
        if status is None:
            raise BaseError("Status is missing")
        logo = json_obj.get('logo')
        return cls(con=con,
                   id=id,
                   shortcode=shortcode,
                   shortname=shortname,
                   longname=longname,
                   description=description,
                   keywords=keywords,
                   ontologies=ontologies,
                   selfjoin=selfjoin,
                   status=status,
                   logo=logo)

    def toJsonObj(self, action: Actions) -> Any:
        """
        Internal method! Should not be used directly!

        Creates a JSON-object from the Project instance that can be used to call Knora

        :param action: Action the object is used for (Action.CREATE or Action.UPDATE)
        :return: JSON-object
        """

        tmp = {}
        if action == Actions.Create:
            if self._shortcode is None:
                raise BaseError("There must be a valid project shortcode!")
            tmp['shortcode'] = self._shortcode
            if self._shortname is None:
                raise BaseError("There must be a valid project shortname!")
            tmp['shortname'] = self._shortname
            if self._longname is None:
                raise BaseError("There must be a valid project longname!")
            tmp['longname'] = self._longname
            if self._description.isEmpty():
                raise BaseError("There must be a valid project description!")
            tmp['description'] = self._description.toJsonObj()
            if self._keywords is not None and len(self._keywords) > 0:
                tmp['keywords'] = self._keywords
            if self._selfjoin is None:
                raise BaseError("selfjoin must be defined (True or False!")
            tmp['selfjoin'] = self._selfjoin
            if self._status is None:
                raise BaseError("status must be defined (True or False!")
            tmp['status'] = self._status
        elif action == Actions.Update:
            if self._shortcode is not None and 'shortcode' in self._changed:
                tmp['shortcode'] = self._shortcode
            if self._shortname is not None  and 'shortname' in self._changed:
                tmp['shortname'] = self._shortname
            if self._longname is not None and 'longname' in self._changed:
                tmp['longname'] = self._longname
            if not self._description.isEmpty() and 'description' in self._changed:
                tmp['description'] = self._description.toJsonObj()
            if len(self._keywords) > 0 and 'keywords' in self._changed:
                tmp['keywords'] = self._keywords
            if self._selfjoin is not None and 'selfjoin' in self._changed:
                tmp['selfjoin'] = self._selfjoin
            if self._status is not None and 'status' in self._changed:
                tmp['status'] = self._status
        return tmp

    def createDefinitionFileObj(self):
        return {
            "shortcode": self._shortcode,
            "shortname": self._shortname,
            "longname": self._longname,
            "descriptions": self._description.createDefinitionFileObj(),
            "keywords": [kw for kw in self._keywords]
        }

    def create(self) -> 'Project':
        """
        Create a new project in Knora

        :return: JSON-object from Knora
        """

        jsonobj = self.toJsonObj(Actions.Create)
        jsondata = json.dumps(jsonobj, cls=SetEncoder)
        result = self._con.post('/admin/projects', jsondata)
        return Project.fromJsonObj(self._con, result['project'])

    def read(self) -> 'Project':
        """
        Read a project from Knora

        :return: JSON-object from Knora
        """
        result = None
        if self._id is not None:
            result = self._con.get('/admin/projects/iri/' + quote_plus(self._id))
        elif self._shortcode is not None:
            result = self._con.get('/admin/projects/shortcode/' + quote_plus(self._shortcode))
        elif self._shortname is not None:
            result = self._con.get('/admin/projects/shortname/' + quote_plus(self._shortname))
        if result is not None:
            return Project.fromJsonObj(self._con, result['project'])
        else:
            return None # Todo: throw exception

    def update(self) -> Union['Project', None]:
        """
        Udate the project info in Knora with the modified data in this project instance

        :return: JSON-object from Knora refecting the update
        """

        jsonobj = self.toJsonObj(Actions.Update)
        if jsonobj:
            jsondata = json.dumps(jsonobj, cls=SetEncoder)
            result = self._con.put('/admin/projects/iri/' + quote_plus(self.id), jsondata)
            return Project.fromJsonObj(self._con, result['project'])
        else:
            return None

    def delete(self) -> 'Project':
        """
        Delete the given Knora project

        :return: Knora response
        """

        result = self._con.delete('/admin/projects/iri/' + quote_plus(self._id))
        return Project.fromJsonObj(self._con, result['project'])

    def set_default_permissions(self, group_id: str) -> None:
        permobj = {
            "forGroup": "http://www.knora.org/ontology/knora-admin#ProjectMember",
            "forProject": self._id,
            "hasPermissions": [
                {
                    "additionalInformation": None,
                    "name": "ProjectResourceCreateAllPermission",
                    "permissionCode": None
                }
            ]
        }
        jsondata = json.dumps(permobj, indent=4)
        print(jsondata)
        result = self._con.post("/admin/permissions/ap", jsondata)
        pprint(result)

        return
        permobj = {
            "forGroup": group_id,
            "forProject": self._id,
            "forProperty": None,
            "forResourceClass": None,
            "hasPermissions": [
                {
                    "additionalInformation": "http://www.knora.org/ontology/knora-admin#ProjectMember",
                    "name": "D",
                    "permissionCode": 7
                }
            ]
        }
        jsondata = json.dumps(permobj)
        result = self._con.post("/admin/permissions/ap", jsondata)
        pprint(result)

    @staticmethod
    def getAllProjects(con: Connection) -> List['Project']:
        """
        Get all existing projects in Knora

        :param con: Connection instance
        :return:
        """
        result = con.get('/admin/projects')
        if 'projects' not in result:
            raise BaseError("Request got no projects!")
        return list(map(lambda a: Project.fromJsonObj(con, a), result['projects']))

    def print(self):
        """
        print info to stdout

        :return: None
        """

        print('Project Info:')
        print('  Id:         {}'.format(self._id))
        print('  Shortcode:  {}'.format(self._shortcode))
        print('  Shortname:  {}'.format(self._shortname))
        print('  Longname:   {}'.format(self._longname))
        if self._description is not None:
            print('  Description:')
            for descr in self._description.items():
                print('    {}: {}'.format(descr[0], descr[1]))
        else:
            print('  Description: None')
        if self._keywords is not None:
            print('  Keywords:   {}'.format(' '.join(self._keywords)))
        else:
            print('  Keywords:   None')
        if self._ontologies is not None:
            print('  Ontologies: {}'.format(' '.join(self._ontologies)))
        else:
            print('  Ontologies: None')
        print('  Selfjoin:   {}'.format(self._selfjoin))
        print('  Status:     {}'.format(self._status))

if __name__ == '__main__':
    con = Connection('http://0.0.0.0:3333')
    con.login('root@example.com', 'test')

    projects = Project.getAllProjects(con)

    for project in projects:
        project.print()

    new_project = Project(con=con,
                               shortcode='F11F',
                               shortname='mytest3',
                               longname='A Test beloning to me',
                               description=LangString({Languages.EN: 'My Tests description'}),
                               keywords={'AAA', 'BBB'},
                               selfjoin=False,
                               status=True).create()

    new_project.print()

    new_project.longname = 'A long name fore a short project'
    new_project.status = False
    new_project.description = LangString({Languages.DE: 'Beschreibung meines Tests'})
    new_project.add_keyword('CCC')
    new_project = new_project.update()
    new_project.print()

    new_project = new_project.delete()

    print('**************************************************************')
    projects = Project.getAllProjects(con)

    for project in projects:
        project.print()



import argparse
import sys
import os
import pkg_resources  # part of setuptools

sys.path.append(os.path.dirname(os.path.realpath(__file__)))

from dsplib.utils.onto_validate import validate_list, validate_ontology
from dsplib.utils.onto_create_lists import create_lists
from dsplib.utils.onto_create_ontology import create_ontology
from dsplib.utils.onto_get import get_ontology
from dsplib.utils.xml_upload import xml_upload
from dsplib.utils.onto_process_excel import list_excel2json


def program(args):
    version = pkg_resources.require("dsp-tools")[0].version

    #
    # parse the arguments of the command line
    #
    parser = argparse.ArgumentParser(
        description=f"dsp-tools (Version {version}) DaSCH Service Platform data modelling tools (© 2021 by DaSCH)."
    )

    subparsers = parser.add_subparsers(title="Subcommands",
                                       description='Valid subcommands are',
                                       help='sub-command help')

    parser_create = subparsers.add_parser('create', help='Create ontologies, lists etc.')
    parser_create.set_defaults(action="create")
    parser_create.add_argument("-s", "--server", type=str, default="http://0.0.0.0:3333", help="URL of the DSP server")
    parser_create.add_argument("-u", "--user", default="root@example.com", help="Username for DSP server")
    parser_create.add_argument("-p", "--password", default="test", help="The password for login")
    parser_create.add_argument("-V", "--validate", action='store_true', help="Do only validation of JSON, no upload of the ontology")
    parser_create.add_argument("-L", "--listfile", type=str, default="lists.json", help="Name of list node informationfile")
    parser_create.add_argument("-l", "--lists", action='store_true', help="Only create the lists")
    parser_create.add_argument("-v", "--verbose", action="store_true", help="Verbose feedback")
    parser_create.add_argument("-d", "--dump", action="store_true", help="dump test files for DSP-API requests")
    parser_create.add_argument("datamodelfile", help="path to data model file")

    parser_get = subparsers.add_parser('get', help='Get project/ontology information from server')
    parser_get.set_defaults(action="get")
    parser_get.add_argument("-u", "--user", default="root@example.com", help="Username for DSP server")
    parser_get.add_argument("-p", "--password", default="test", help="The password for login")
    parser_get.add_argument("-s", "--server", type=str, default="http://0.0.0.0:3333", help="URL of the DSP server")
    parser_get.add_argument("-P", "--project", type=str, help="Shortcode, shortname or iri of project", required=True)
    parser_get.add_argument("-v", "--verbose", action="store_true", help="Verbose feedback")
    parser_get.add_argument("datamodelfile", help="path to data model file", default="onto.json")

    parser_upload = subparsers.add_parser('xmlupload', help='Upload data from XML file to server')
    parser_upload.set_defaults(action="xmlupload")
    parser_upload.add_argument("-s", "--server", type=str, default="http://0.0.0.0:3333", help="URL of the DSP server")
    parser_upload.add_argument("-u", "--user", type=str, default="root@example.com", help="Username for DSP server")
    parser_upload.add_argument("-p", "--password", type=str, default="test", help="The password for login")
    parser_upload.add_argument("-V", "--validate", action='store_true', help="Do only validation of JSON, no upload of the ontology")
    parser_upload.add_argument("-i", "--imgdir", type=str, default=".", help="Path to folder containing the images")
    parser_upload.add_argument("-S", "--sipi", type=str, default="http://0.0.0.0:1024", help="URL of SIPI server")
    parser_upload.add_argument("-v", "--verbose", action="store_true", help="Verbose feedback")
    parser_upload.add_argument("xmlfile", help="path to xml file containing the data", default="data.xml")

    parser_excellists = subparsers.add_parser('excel', help='Create lists JSON from excel files')
    parser_excellists.set_defaults(action="excel")
    parser_excellists.add_argument("-S", "--sheet", type=str, help="Name of excel sheet to be used", default="Tabelle1")
    parser_excellists.add_argument("-s", "--shortcode", type=str, help="Shortcode of project", default="4123")
    parser_excellists.add_argument("-l", "--listname", type=str, help="Name of list to be created", default="my_list")
    parser_excellists.add_argument("-L", "--label", type=str, help="Label of list to be created", default="MyList")
    parser_excellists.add_argument("-x", "--lang", type=str, help="Language for label", default="en")
    parser_excellists.add_argument("-v", "--verbose", action="store_true", help="Verbose feedback")
    parser_excellists.add_argument("excelfile", help="Path to the excel file containing the list data", default="lists.xlsx")
    parser_excellists.add_argument("outfile", help="Path to the output JSON file containing the list data", default="list.json")

    args = parser.parse_args(args)

    if not hasattr(args, 'action'):
        parser.print_help(sys.stderr)
        exit(0)

    if args.action == "create":
        if args.lists:
            if args.validate:
                validate_list(args.datamodelfile)
            else:
                create_lists(args.datamodelfile, args.dump)
        else:
            if args.validate:
                validate_ontology(args.datamodelfile)
            else:
                create_ontology(input_file=args.datamodelfile,
                                lists_file=args.listfile,
                                server=args.server,
                                user=args.user,
                                password=args.password,
                                verbose=args.verbose,
                                dump=args.dump if args.dump else False)
    elif args.action == "get":
        get_ontology(projident=args.project,
                     outfile=args.datamodelfile,
                     server=args.server,
                     user=args.user,
                     password=args.password,
                     verbose=args.verbose)
    elif args.action == "xmlupload":
        xml_upload(input_file=args.xmlfile,
                   server=args.server,
                   user=args.user,
                   password=args.password,
                   imgdir=args.imgdir,
                   sipi=args.sipi,
                   verbose=args.verbose,
                   validate=args.validate)
    elif args.action == "excel":
        list_excel2json(excelpath=args.excelfile,
                        sheetname=args.sheet,
                        shortcode=args.shortcode,
                        listname=args.listname,
                        label=args.label,
                        lang=args.lang,
                        outfile=args.outfile,
                        verbose=args.verbose)



def main():
    program(sys.argv[1:])


if __name__ == '__main__':
    program(sys.argv[1:])

